'''Plot univariate data (:class:`Var` objects).

.. autosummary::
   :toctree: generated

   Barplot
   boxcox_explore
   Boxplot
   Correlation
   Histogram
   MultiTimeplot
   Regression
   Timeplot

'''
# author: Christian Brodbeck

from __future__ import division

from itertools import izip
import logging

import numpy as np
import scipy.stats  # without this sp.stats is not available
import matplotlib.pyplot as plt
import matplotlib as mpl

from .._stats import test
from .._data_obj import (asfactor, isvar, asvar, ascategorial, assub, cellname,
                         Celltable)
from ._base import _EelFigure, str2tex


defaults = dict(title_kwargs={'size': 14,
                                'family': 'serif'},
              mono=False,  # use monochrome instead of colors
              # defaults for color
              hatch=['', '', '//', '', '*', 'O', '.', '', '/', '//'],
              linestyle=['-', '-', '--', ':'],
              c={'pw': ['#00FF00', '#FFCC00', '#FF6600', '#FF3300'],
#                   'colors': ['.3', '.7', '1.', '1.'],
                   'colors': [(0.99609375, 0.12890625, 0.0),
                              (0.99609375, 0.5859375, 0.0),
                              (0.98046875, 0.99609375, 0.0),
                              (0.19921875, 0.99609375, 0.0)],
                   'markers': ['o', 'o', '^', 'o'],
                   },
              # defaults for monochrome
              cm={'pw': ['.6', '.4', '.2', '.0'],
                    'colors': ['.3', '.7', '1.', '1.'],
                    'markers': ['o', 'o', '^', 'o'],
                    },
                )  # set by __init__








def _mark_plot_pairwise(ax, ct, par, y_min, y_unit, x0=0,
                        corr='Hochberg', levels=True, trend=".", pwcolors=None,
                        font_size=plt.rcParams['font.size'] * 1.5
                        ):
    "returns y_max"
    if levels is not True:  # to avoid test.star() conflict
        trend = False
    # tests
    if not pwcolors:
        if defaults['mono']:
            pwcolors = defaults['cm']['pw'][1 - bool(trend):]
        else:
            pwcolors = defaults['c']['pw'][1 - bool(trend):]
    k = len(ct.cells)
    tests = test._pairwise(ct.get_data(), within=ct.all_within, parametric=par, trend=trend,
                            levels=levels, corr=corr)
    reservation = np.zeros((sum(xrange(1, k)), k - 1))
    y_top = y_min  # track top of plot
    y_start = y_min + 2 * y_unit
    # loop through possible connections
    for distance in xrange(1, k):
        for i in xrange(0, k - distance):
            j = i + distance  # i, j are data indexes for the categories being compared
            index = tests['pw_indexes'][(i, j)]
            stars = tests['stars'][index]
            if stars:
                c = pwcolors[stars - 1]
                symbol = tests['symbols'][index]

                free_levels = np.where(reservation[:, i:j].sum(1) == 0)[0]
                level = min(free_levels)
                reservation[level, i:j] = 1

                y1 = y_start + 2 * y_unit * level
                y2 = y1 + y_unit
                y_top = max(y2, y_top)
                x1 = (x0 + i) + .025
                x2 = (x0 + j) - .025
                ax.plot([x1, x1, x2, x2], [y1, y2, y2, y1], color=c)
                ax.text((x1 + x2) / 2, y2, symbol, color=c, size=font_size,
                        horizontalalignment='center', clip_on=False,
                        verticalalignment='center', backgroundcolor='w')
    y_top = y_top + 2 * y_unit
    return y_top

def _mark_plot_1sample(ax, ct, par, y_min, y_unit, x0=0, corr='Hochberg',
                        levels=True, trend=".", pwcolors=None,
                        popmean=0,  # <- mod
                        font_size=plt.rcParams['font.size'] * 1.5
                        ):
    "returns y_max"
    if levels is not True:  # to avoid test.star() conflict
        trend = False
    # tests
    if not pwcolors:
        if defaults['mono']:
            pwcolors = defaults['cm']['pw'][1 - bool(trend):]
        else:
            pwcolors = defaults['c']['pw'][1 - bool(trend):]
    # mod
    ps = []
    if par:
        for d in ct.get_data():
            t, p = scipy.stats.ttest_1samp(d, popmean)
            ps.append(p)
    else:
        raise NotImplementedError("nonparametric 1-sample test")
    stars = test.star(ps, int, levels, trend, corr)
    stars_str = test.star(ps, levels=levels, trend=trend)
    if any(stars):
        y_stars = y_min + 1.75 * y_unit
        for i, n_stars in enumerate(stars):
            if n_stars > 0:
                c = pwcolors[n_stars - 1]
                ax.text(x0 + i, y_stars, stars_str[i], color=c, size=font_size,
                        horizontalalignment='center', clip_on=False,
                        verticalalignment='center')
        return y_min + 4 * y_unit
    else:
        return y_min


class _SimpleFigure(_EelFigure):
    def __init__(self, wintitle, title=None, xlabel=None, ylabel=None,
                 titlekwargs=defaults['title_kwargs'], yticks=None,
                 figsize=(5, 5), xtick_rotation=0, ytick_rotation=0,
                 frame=True):
        _EelFigure.__init__(self, wintitle, fig_kwa={'figsize': figsize})

        # axes
        ax_x0 = .025 + .07 * bool(ylabel)
        ax_y0 = .065 + .055 * bool(xlabel)
        ax_dx = .975 - ax_x0
        ax_dy = .95 - ax_y0 - .08 * bool(title)
        rect = [ax_x0, ax_y0, ax_dx, ax_dy]
        ax = self.figure.add_axes(rect)
        if not frame:
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ax.yaxis.set_ticks_position('left')
            ax.xaxis.set_ticks_position('bottom')
        self._axes.append(ax)
        self._ax = ax

        # ticks / tick labels
        self._yticks = yticks
        self._x_tick_rotation = xtick_rotation
        self._y_tick_rotation = ytick_rotation
        xax = ax.get_xaxis()
        xax.set_ticks_position('none')

        # collector for handles for figure legend
        self._handles = []
        self._legend = None

        # title and labels
        if title:
            if 'verticalalignment' not in titlekwargs:
                titlekwargs['verticalalignment'] = 'bottom'
            ax.set_title(title, **titlekwargs)
        if ylabel:
            ax.set_ylabel(ylabel)  # , labelpad=-20.)
        if xlabel:
            ax.set_xlabel(xlabel)

    def add_legend_handles(self, *handles):
        for handle in handles:
            label = handle.get_label()
            if not self.legend_has_label(label):
                self._handles.append(handle)

    def legend_has_label(self, label):
            return any(label == h.get_label() for h in self._handles)

    def legend(self, loc=0, fig=False, zorder=-1, ncol=1):
        "add a legend to the plot"
        if fig:
            labels = (h.get_label() for h in self._handles)
            l = self.figure.legend(self._handles, labels, loc, ncol=ncol)
            self._legend = l
        else:
            l = self._ax.legend(loc=loc, ncol=ncol)
            if l:
                l.set_zorder(-1)
            else:
                raise ValueError("No labeled plot elements for legend")

    def _show(self):
        "resizes the axes to take into account tick spacing"

        if self._yticks:
            yticks = self._yticks
            if np.iterable(yticks[0]):
                locations, labels = yticks
            else:
                locations = yticks
                labels = None
            self._ax.set_yticks(locations)
            if labels:
                self._ax.set_yticklabels(labels)
        if self._x_tick_rotation:
            for t in self._ax.get_xticklabels():
                t.set_rotation(self._x_tick_rotation)
        if self._y_tick_rotation:
            for t in self._ax.get_yticklabels():
                t.set_rotation(self._y_tick_rotation)

        # adjust the position of the axes to show all labels
        self.draw()
        x_in, y_in = self.figure.get_size_inches()
        dpi = self.figure.get_dpi()
        border_x0 = 0.05  # in inches
        border_x1 = 0.05  # in inches
        border_y0 = 0.05  # in inches
        border_y1 = 0.05  # in inches
        if self._legend:
            w = self._legend.get_window_extent()
            border_y0 += w.ymax / dpi

        xmin = x_in * dpi
        ymin = y_in * dpi
        xmax = 0
        ymax = 0
        for c in self._ax.get_children():
            try:
                w = c.get_window_extent()
            except:
                pass
            else:
                xmin = min(xmin, w.xmin)
                ymin = min(ymin, w.ymin)
                xmax = max(xmax, w.xmax)
                ymax = max(ymax, w.ymax)

        for label in self._ax.get_ymajorticklabels() + \
                     [self._ax.get_yaxis().get_label()]:
            w = label.get_window_extent()
            xmin = min(xmin, w.xmin)

        for label in self._ax.get_xmajorticklabels() + \
                     [self._ax.get_xaxis().get_label()]:
            w = label.get_window_extent()
            ymin = min(ymin, w.ymin)

        # to figure proportion
        xmin = (xmin / dpi - border_x0) / x_in
        xmax = (xmax / dpi + border_x1 - x_in) / x_in
        ymin = (ymin / dpi - border_y0) / y_in
        ymax = (ymax / dpi + border_y1 - y_in) / y_in

        p = self._ax.get_position()
        p.x0 -= xmin
        p.x1 -= xmax
        p.y0 -= ymin
        p.y1 -= ymax
        self._ax.set_position(p)

        _EelFigure._show(self, False)



class Boxplot(_SimpleFigure):
    def __init__(self, Y, X=None, match=None, sub=None, datalabels=None,
                 bottom=None, top=None,
                 title=True, ylabel='{unit}', xlabel=True, xtick_delim='\n',
                 titlekwargs=defaults['title_kwargs'],
                 test=True, par=True, trend=".", corr='Hochberg',
                 pwcolors=None, hatch=False, colors=False,
                 ds=None, **kwargs):
        """
        Make a boxplot.

        Parameters
        ----------
        Y : Var
            Dependent variable.
        X : categorial
            Category definition (draw one box for every cell in X).
        match : None | categorial
            Match cases for a repeated measures test.
        sub : None | index-array
            Perform test with a subset of the data.
        datalabels : scalar
            Threshold for labeling outliers (in standard-deviation).
        bottom : scalar
            Lowest possible value on the y axis (default is 0 or slightly
            below the lowest value).
        top : scalar
            Set the upper x axis limit (default is to fit all the data).
        title : None | str
            Axis title (default is ``Y.name``).
        ylabel : None | str
            Y axis label (default is ``Y.info['unit']`` if present, otherwise
            no label).
        xlabel : None | str
            X axis label (default is ``X.name``).
        xtick_delim : str
            Delimiter for x axis category descriptors (default is ``'\n'``,
            i.e. the level on each Factor of ``X`` on a separate line).
        titlekwargs : dict
            Keyword arguments for the figure title.
        test : bool | scalar
            True (default): perform pairwise tests;  False/None: no tests;
            scalar: 1-sample tests against this value.
        par : bool
            Use parametric test for pairwise comparisons (use non-parametric
            tests if False).
        trend : None | str
            Marker for a trend in pairwise comparisons.
        corr : None | 'hochberg' | 'bonferroni' | 'holm'
            Method for multiple comparison correction (default 'hochberg').
        hatch : bool | str
            Matplotlib Hatch pattern to fill boxes (True to use the module
            default; default is False).
        colors : bool | sequence of matplitlib colors
            Matplotlib colors to use for boxes (True to use the module default;
            default is False, i.e. no colors).
        ds : None | Dataset
            If a Dataset is specified, all data-objects can be specified as
            names of Dataset variables
        frame : bool
            Draw a frame containing the figure from the top and the right
            (default ``True``).
        """
        # get data
        ct = Celltable(Y, X, match=match, sub=sub, ds=ds, coercion=asvar)
        Y = ct.Y
        X = ct.X

        # kwargs
        if hatch == True:
            hatch = defaults['hatch']
        if colors == True:
            if defaults['mono']:
                colors = defaults['cm']['colors']
            else:
                colors = defaults['c']['colors']

        if title is True:
            title = str2tex(getattr(Y, 'name', None))

        # ylabel
        if hasattr(Y, 'info'):
            unit = Y.info.get('unit', '')
        else:
            unit = ''
        ylabel = ylabel.format(unit=unit)

        # xlabel
        if xlabel is True:
            xlabel = str2tex(X.name or False)

        # get axes


        _SimpleFigure.__init__(self, "Boxplot", title, xlabel, ylabel,
                               titlekwargs, **kwargs)
        ax = self._axes[0]

        # determine ax lim
        if bottom is None:
            if np.min(ct.Y.x) >= 0:
                bottom = 0
            else:
                d_min = np.min(ct.Y.x)
                d_max = np.max(ct.Y.x)
                d_range = d_max - d_min
                bottom = d_min - .05 * d_range

        # boxplot
        k = len(ct.cells)
        all_data = ct.get_data()
        bp = ax.boxplot(all_data)

        # Now fill the boxes with desired colors
        if hatch or colors:
            numBoxes = len(bp['boxes'])
            for i in range(numBoxes):
                box = bp['boxes'][i]
                boxX = box.get_xdata()[:5]  # []
                boxY = box.get_ydata()[:5]  # []
                boxCoords = zip(boxX, boxY)
                # Alternate between Dark Khaki and Royal Blue
                if len(colors) >= numBoxes:
                    c = colors[i]
                else:
                    c = '.5'
                if len(hatch) >= numBoxes:
                    h = hatch[i]
                else:
                    h = ''
                boxPolygon = mpl.patches.Polygon(boxCoords, facecolor=c, hatch=h, zorder=-999)
                ax.add_patch(boxPolygon)
        if defaults['mono']:
            for itemname in bp:
                plt.setp(bp[itemname], color='black')

        # labelling
        ax.set_xticks(np.arange(len(ct.cells)) + 1)
        ax.set_xticklabels(ct.cellnames(xtick_delim))
        y_min = np.max(np.hstack(all_data))
        y_unit = (y_min - bottom) / 15

        # tests
        if test is True:
            y_top = _mark_plot_pairwise(ax, ct, par, y_min, y_unit, corr=corr,
                                        x0=1, trend=trend)
        else:
            ax.axhline(test, color='black')
            y_top = _mark_plot_1sample(ax, ct, par, y_min, y_unit,
                                       x0=1, corr=corr, popmean=test, trend=trend)
        if top is None:
            top = y_top

        # data labels
        if datalabels:
            for i, cell in enumerate(ct.cells):
                d = ct.data[cell]
                indexes = np.where(np.abs(d) / d.std() >= datalabels)[0]
                for index in indexes:
                    label = ct.matchlabels[cell][index]
                    ax.annotate(label, (i + 1, d[index]))

        # set ax limits
        ax.set_ylim(bottom, top)
        ax.set_xlim(.5, k + .5)

        self._show()


class Barplot(_SimpleFigure):
    def __init__(self, Y, X=None, match=None, sub=None, test=True, par=True,
                 corr='Hochberg', title=True, trend=".",
                 # bar settings:
                 ylabel='{unit}{err}', err='2sem', ec='k', xlabel=True,
                 xtick_delim='\n', hatch=False, colors=False, bottom=0,
                 c='#0099FF', edgec=None, ds=None, **kwargs):
        """Barplot

        Parameters
        ----------
        Y : scalar Var
            Dependent variable.
        X : categorial
            Model (Factor or Interaction)
        match : None | categorial
            Match cases for a repeated measures test.
        sub : None | index-array
            Perform test with a subset of the data.
        test : bool | scalar
            True (default): perform pairwise tests;  False: no tests;
            scalar: 1-sample tests against this value
        par : bool
            Use parametric test for pairwise comparisons (use non-parametric
            tests if False).
        corr : None | 'hochberg' | 'bonferroni' | 'holm'
            Method for multiple comparison correction (default 'hochberg').
        title : None | str
            Axis title (default is ``Y.name``).
        trend : None | str
            Marker for a trend in pairwise comparisons.
        ylabel : None | str
            Y axis label (default is ``Y.info['unit']`` if present, followed
            by the error bar description).
        err: func | "[x][type]"
            The magnitude of the error bars to display. Examples:
            'std' : 1 standard deviation;
            '.95ci' : 95% confidence interval (see :func:`..ci`);
            '2sem' : 2 standard error of the mean
        ec : matplotlib color
            Error bar color.
        xlabel : None | str
            X axis label (default is ``X.name``).
        xtick_delim : str
            Delimiter for x axis category descriptors (default is ``'\n'``,
            i.e. the level on each Factor of ``X`` on a separate line).
        hatch : bool | str
            Matplotlib Hatch pattern to fill boxes (True to use the module
            default; default is False).
        colors : bool | sequence of matplitlib colors
            Matplotlib colors to use for boxes (True to use the module default;
            default is False, i.e. no colors).
        bottom : scalar
            Lowest possible value on the y axis (default is 0).
        c : matplotlib color
            Bar color (ignored if colors is specified).
        edgec : matplotlib color
            Barplot edge color.
        ds : None | Dataset
            If a Dataset is specified, all data-objects can be specified as
            names of Dataset variables
        frame : bool
            Draw a frame containing the figure from the top and the right
            (default ``True``).
        """
        if title is True:
            title = getattr(Y, 'name', None)

        if isinstance(err, basestring):
            if err.endswith('ci'):
                if len(err) > 2:
                    a = float(err[:-2])
                else:
                    a = .95
                err_desc = '$\pm %g ci$' % a
            elif err.endswith('sem'):
                if len(err) > 3:
                    a = float(err[:-3])
                else:
                    a = 1
                err_desc = '$\pm %g sem$' % a
            elif err.endswith('std'):
                if len(err) > 3:
                    a = float(err[:-3])
                else:
                    a = 1
                err_desc = '$\pm %g std$' % a
            else:
                raise ValueError('unrecognized statistic: %r' % err)
        else:
            err_desc = getattr(err, '__name__', '')

        # ylabel
        if ylabel:
            if hasattr(Y, 'info'):
                unit = Y.info.get('unit', '')
            else:
                unit = ''

            if ylabel is True:
                if unit:
                    ylabel = unit
                else:
                    ylabel = False
            elif isinstance(ylabel, str):
                ylabel = ylabel.format(unit=unit, err=err_desc)

        # xlabel
        if xlabel is True:
            if hasattr(X, 'name'):
                xlabel = X.name.replace('_', ' ')
            else:
                xlabel = False

        _SimpleFigure.__init__(self, "BarPlot", title, xlabel, ylabel, **kwargs)

        ct = Celltable(Y, X, match=match, sub=sub, ds=ds, coercion=asvar)

        x0, x1, y0, y1 = _barplot(self._ax, ct,
                                  test=test, par=par, trend=trend, corr=corr,
                                  # bar settings:
                                  err=err, ec=ec,
                                  hatch=hatch, colors=colors,
                                  bottom=bottom, c=c, edgec=edgec,
                                  return_lim=True)

        self._ax.set_xlim(x0, x1)
        self._ax.set_ylim(y0, y1)

        # figure decoration
        self._ax.set_xticks(np.arange(len(ct.cells)))
        self._ax.set_xticklabels(ct.cellnames(xtick_delim))

        self._show()


def _barplot(ax, ct,
             test=True, par=True, trend=".", corr='Hochberg',
             # bar settings:
             err='2sem', ec='k',
             hatch=False, colors=False,
             bottom=0, c='#0099FF', edgec=None,
             left=None, width=.5,
             return_lim=False,
             ):
    """
    draw a barplot to axes ax for Celltable ct.

    return_lim: return axes limits (x0, x1, y0, y1)
    """
    # kwargs
    if hatch == True:
        hatch = defaults['hatch']
    if colors == True:
        if defaults['mono']:
            colors = defaults['cm']['colors']
        else:
            colors = defaults['c']['colors']
    # data
    k = len(ct.cells)
    if left is None:
        left = np.arange(k) - width / 2
    height = np.array(ct.get_statistic(np.mean))
    y_error = np.array(ct.get_statistic(err))


    # fig spacing
    plot_max = np.max(height + y_error)
    plot_min = np.min(height - y_error)
    plot_span = plot_max - plot_min
    y_bottom = min(bottom, plot_min - plot_span * .05)

    # main BARPLOT
    bars = ax.bar(left, height - bottom, width, bottom=bottom,
                  color=c, edgecolor=edgec, linewidth=1,
                  ecolor=ec, yerr=y_error)

    # hatch
    if hatch:
        for bar, h in zip(bars, hatch):
            bar.set_hatch(h)
    if colors:
        for bar, c in zip(bars, colors):
            bar.set_facecolor(c)

    # pairwise tests
    # prepare pairwise plotting
    if y_error == None:
        y_min = np.max(height)
    else:
        y_min = np.max(height + y_error)
    y_unit = (y_min - y_bottom) / 15
    if test is True:
        y_top = _mark_plot_pairwise(ax, ct, par, y_min, y_unit,
                                    corr=corr, trend=trend)
    elif (test is False) or (test is None):
        y_top = y_min + y_unit
    else:
        ax.axhline(test, color='black')
        y_top = _mark_plot_1sample(ax, ct, par, y_min, y_unit,
                                   popmean=test, corr=corr, trend=trend)

    #      x0,                 x1,                  y0,       y1
    if return_lim:
        lim = (min(left) - .5 * width, max(left) + 1.5 * width, y_bottom, y_top)
        return lim


class Timeplot(_SimpleFigure):
    def __init__(self, Y, categories, time, match=None, sub=None, ds=None,
                 # data plotting
                 main=np.mean,
                 spread='box', x_jitter=False,
                 datalabels=None,
                 # labelling
                 ylabel=True, xlabel=True,
                 legend=True, loc=0,
                 colors=True, hatch=False, markers=True,
                 **kwargs
                 ):
        """Plot a variable over time.

        Parameters
        ----------
        spread : str
            How to indicator data spread.
            None - without
            'box' - boxplots
            for lineplots:
            'Xsem' X standard error of the means
            'Xstd' X standard deviations
        main : numpy function
            draw lines to connect values across time (default: np.mean)
            can be 'bar' for barplots or False
        datalabels : scalar
            Label outlier data. The argument is provided in std.
        diff : scalar
            Use this value as baseline for plotting; test other conditions
            agaist baseline (instead of pairwise)
        legend : bool | 'fig'
            Plot a legend; with `fig`, plot as figlegend.
        loc :
            The legend location.
        frame : bool
            Draw a frame containing the figure from the top and the right
            (default ``True``).
        """
        sub = assub(sub, ds)
        Y = asvar(Y, sub, ds)
        categories = ascategorial(categories, sub, ds)
        time = asvar(time, sub, ds)
        if match is not None:
            match = ascategorial(match, sub, ds)

        # transform to 3 kwargs:
        # - local_plot ('bar' or 'box')
        # - line_plot (function for values to connect)
        # - spread
        if main == 'bar':
            assert spread != 'box'
            local_plot = 'bar'
            line_plot = None
        else:
            line_plot = main
            if spread == 'box':
                local_plot = 'box'
                spread = None
            else:
                local_plot = None
        del main

        # hatch/marker
        if hatch == True:
            hatch = defaults['hatch']
        if markers == True:
            if defaults['mono']:
                markers = defaults['cm']['markers']
            else:
                markers = defaults['c']['markers']

        # colors: {category index -> color, ...}
        colors_arg = colors
        if defaults['mono']:
            colors = dict(zip(categories.cells, defaults['cm']['colors']))
        else:
            colors = dict(zip(categories.cells, defaults['c']['colors']))
        if isinstance(colors_arg, dict):
            colors.update(colors_arg)
        elif isinstance(colors_arg, (tuple, list)):
            colors.update(dict(zip(categories.cells, colors_arg)))

        for c in categories.cells:
            if c not in colors:
                colors[c] = '.5'

        color_list = [colors[i] for i in sorted(colors.keys())]

        # ylabel
        if ylabel is True:
            ylabel = str2tex(getattr(Y, 'name', None))

        # xlabel
        if xlabel is True:
            xlabel = str2tex(time.name)

        # get axes
        _SimpleFigure.__init__(self, "Timeplot", xlabel=xlabel, ylabel=ylabel,
                               **kwargs)
        ax = self._ax

        # categories
        n_cat = len(categories.cells)

        # find time points
        if isvar(time):
            time_points = np.unique(time.x)
            n_time_points = len(time_points)
            time_step = min(np.diff(time_points))
            if local_plot in ['box', 'bar']:
                within_spacing = time_step / (2 * n_cat)
                padding = (2 + n_cat / 2) * within_spacing
            else:
                within_spacing = time_step / (8 * n_cat)
                padding = time_step / 4  # (2 + n_cat/2) * within_spacing

            rel_pos = np.arange(0, n_cat * within_spacing, within_spacing)
            rel_pos -= np.mean(rel_pos)

            t_min = min(time_points) - padding
            t_max = max(time_points) + padding
        else:
            raise NotImplementedError("time needs to be Var object")

        # prepare array for timelines
        if line_plot:
            line_values = np.empty((n_cat, n_time_points))
        if spread  and  local_plot != 'bar':
            yerr = np.empty((n_cat, n_time_points))

        # loop through time points
        for i_t, t in enumerate(time_points):
            ct = Celltable(Y, categories, match=match, sub=(time == t))
            if line_plot:
                line_values[:, i_t] = ct.get_statistic(line_plot)

            pos = rel_pos + t
            if local_plot == 'box':
                # boxplot
                bp = ax.boxplot(ct.get_data(out=list), positions=pos, widths=within_spacing)

                # Now fill the boxes with desired colors
#               if hatch or colors:
#                   numBoxes = len(bp['boxes'])
                for i, cell in enumerate(ct.cells):
                    box = bp['boxes'][i]
                    boxX = box.get_xdata()[:5]
                    boxY = box.get_ydata()[:5]
                    boxCoords = zip(boxX, boxY)

                    c = colors[cell]
                    try:
                        h = hatch[i]
                    except:
                        h = ''
                    boxPolygon = mpl.patches.Polygon(boxCoords, facecolor=c, hatch=h, zorder=-999)
                    ax.add_patch(boxPolygon)

                if True:  # defaults['mono']:
                    for itemname in bp:
                        plt.setp(bp[itemname], color='black')
            elif local_plot == 'bar':
                lim = _barplot(ax, ct, test=False, err=spread,  # ec=ec,
                               # bar settings:
                               hatch=hatch, colors=color_list,
                               bottom=0, left=pos, width=within_spacing)
            elif spread:
                yerr[:, i_t] = ct.get_statistic(spread)

        if line_plot:
            # plot means
            x = time_points
            for i, cell in enumerate(categories.cells):
                y = line_values[i]
                name = cellname(cell)
                color = colors[cell]

                if hatch:
                    ls = defaults['linestyle'][i]
                    if color == '1.':
                        color = '0.'
                else:
                    ls = '-'

                if ls == '-':
                    mfc = color
                else:
                    mfc = '1.'

                try:
                    marker = markers[i]
                except:
                    marker = None

                handles = ax.plot(x, y, color=color, linestyle=ls, label=name,
                                  zorder=6, marker=marker, mfc=mfc)
                self.add_legend_handles(*handles)

                if spread:
                    if x_jitter:
                        x_errbars = x + rel_pos[i]
                    else:
                        x_errbars = x
                    ax.errorbar(x_errbars, y, yerr=yerr[i], fmt=None, zorder=5,
                                ecolor=color, linestyle=ls, label=name)
        else:
            legend = False

        # finalize
        ax.set_xlim(t_min, t_max)
        ax.set_xticks(time_points)

        if any (legend is i for i in (False, None)):
            pass
        elif legend == 'fig':
            self.figure.legend(fig=True, loc='lower center')
        else:
            if legend is True:
                loc = 0
            else:
                loc = legend
            self.legend(loc=loc)

        self._show()


class MultiTimeplot(_SimpleFigure):
    def __init__(self, figsize=(7, 2.5),
                 tpad=.5, ylim=None,
                 main=np.mean,
                 spread='box', x_jitter=False,
                 datalabels=None,
                 # labelling
                 title=None, ylabel=True, xlabel=True,
                 titlekwargs=defaults['title_kwargs'],
                 ):
        """
        Template for a figure including multiple timeplots.

        Create an empty template figure for a plot subsuming several
        :func:`timeplot`s. Add timeplots using the ``plot`` method.


        Parameters
        ----------

        ylim : None | (bottom, top) tuple of scalars
            By default (None), ylim depend on the data plotted.

        """
        self._ylim = ylim
        # ylabel
        if ylabel is True:
            self._ylabel = True
            ylabel = False
        else:
            self._ylabel = False
        # xlabel
        if xlabel is True:
            self._xlabel = True
            xlabel = False
        else:
            self._xlabel = False

        # transform to 3 kwargs:
        # - local_plot ('bar' or 'box')
        # - line_plot (function for values to connect)
        # - spread
        if main == 'bar':
            assert spread != 'box'
            local_plot = 'bar'
            line_plot = None
        else:
            line_plot = main
            if spread == 'box':
                local_plot = 'box'
                spread = None
            else:
                local_plot = None
        del main
        self._local_plot = local_plot
        self._line_plot = line_plot
        self._spread = spread
        self._x_jitter = x_jitter


        self._tstart = None
        self._tend = None
        self._tpad = tpad

        self._xticks = []
        self._xticklabels = []

        self._headings = []  # collect text objects for headings to adjust y
        # [(forced_y, line, text), ...]
        self._heading_y = None  # used heading y
        self._y_unit = 0

        # get axes
        _SimpleFigure.__init__(self, title, xlabel, ylabel, titlekwargs,
                               figsize=figsize)

    def plot(self, Y, categories, time, match=None, sub=None,
             tskip=1,
             heading=None, headingy=None,
             colors=True, hatch=False, markers=True):
        """
        Parameters
        ----------

        Y :
            variable to plot
        categories :
            Factor indicating different categories to plot
        time :
            variable indicating time
        match :
            Factor which indicates dependent measures (e.g. subject)
        sub :
            subset
        heading :
            heading for this set (if None, no heading is added)
        headingy :
            y coordinate on which to plot the heading


        See also timeplot parameters.

        """
        categories = asfactor(categories)

        ax = self._ax
        local_plot = self._local_plot
        line_plot = self._line_plot
        spread = self._spread

        # labels
        if self._ylabel is True:
            ylabel = str2tex(getattr(Y, 'name', None))
            if ylabel:
                ax.set_ylabel(ylabel)
                self._ylabel = False
        if self._xlabel is True:
            ax.set_xlabel(str2tex(time.name))
            self._xlabel = False

        ### same as timeplot() #####  #####  #####  #####  #####  #####
        # hatch/marker
        if hatch == True:
            hatch = defaults['hatch']
        if markers == True:
            if defaults['mono']:
                markers = defaults['cm']['markers']
            else:
                markers = defaults['c']['markers']

        # colors: {category index -> color, ...}
        colors_arg = colors
        if defaults['mono']:
            colors = dict(zip(categories.indexes, defaults['cm']['colors']))
        else:
            colors = dict(zip(categories.indexes, defaults['c']['colors']))
        colors.update(categories.colors)
        if isinstance(colors_arg, dict):
            colors.update(colors_arg)
        elif isinstance(colors_arg, (tuple, list)):
            colors.update(dict(zip(categories.indexes, colors_arg)))

        for c in categories.cells:
            if c not in colors:
                colors[c] = '.5'

        color_list = [colors[i] for i in sorted(colors.keys())]


        # sub
        if sub is not None:
            Y = Y[sub]
            categories = categories[sub]
            time = time[sub]
            match = match[sub]

        # categories
        n_cat = len(categories.cells)

        # find time points
        if isvar(time):
            time_points = np.unique(time.x)
        ### NOT same as timeplot() #####  #####  #####  #####  #####  #####
            if self._tend is None:
                t_add = 0
                self._tstart = time_points[0]
                self._tend = time_points[-1]
            else:
                t_add = self._tend + tskip - time_points[0]
                self._tend = t_add + time_points[-1]
        ### same as timeplot()     #####  #####  #####  #####  #####  #####
            n_time_points = len(time_points)
            time_step = min(np.diff(time_points))
            if local_plot in ['box', 'bar']:
                within_spacing = time_step / (2 * n_cat)
            else:
                within_spacing = time_step / (8 * n_cat)

            rel_pos = np.arange(0, n_cat * within_spacing, within_spacing)
            rel_pos -= np.mean(rel_pos)

        ### NOT same as timeplot() #####  #####  #####  #####  #####  #####
            t_min = self._tstart - self._tpad
            t_max = self._tend + self._tpad
        ### same as timeplot()     #####  #####  #####  #####  #####  #####

        else:
            raise NotImplementedError("time needs to be Var object")

        # prepare array for timelines
        if line_plot:
            line_values = np.empty((n_cat, n_time_points))
        if spread  and  local_plot != 'bar':
            yerr = np.empty((n_cat, n_time_points))

        # loop through time points
        ymax = None; ymin = None
        for i_t, t in enumerate(time_points):
            ct = Celltable(Y, categories, match=match, sub=(time == t))
            if line_plot:
                line_values[:, i_t] = ct.get_statistic(line_plot)

            pos = rel_pos + t
            if local_plot == 'box':
                # boxplot
                data = ct.get_data(out=list)
                bp = ax.boxplot(data, positions=pos, widths=within_spacing)

                ymax_loc = np.max(data)
                ymin_loc = np.min(data)

                # Now fill the boxes with desired colors
                for i, cat in enumerate(ct.indexes):
                    box = bp['boxes'][i]
                    boxX = box.get_xdata()[:5]
                    boxY = box.get_ydata()[:5]
                    boxCoords = zip(boxX, boxY)

                    c = colors[cat]
                    try:
                        h = hatch[i]
                    except:
                        h = ''
                    boxPolygon = mpl.patches.Polygon(boxCoords, facecolor=c, hatch=h, zorder=-999)
                    ax.add_patch(boxPolygon)

                if True:  # defaults['mono']:
                    for itemname in bp:
                        plt.setp(bp[itemname], color='black')
            elif local_plot == 'bar':
                lim = _barplot(ax, ct, test=False, err=spread,  # ec=ec,
                               # bar settings:
                               hatch=hatch, colors=color_list,
                               bottom=0, left=pos, width=within_spacing)
                ymax_loc = lim[1]
                ymin_loc = lim[0]
            elif spread:
                yerr_loc = np.array(ct.get_statistic(spread))
                yerr[:, i_t] = yerr_loc
                y_loc = np.array(ct.get_statistic(np.mean))
                ymax_loc = max(y_loc + yerr_loc)
                ymin_loc = min(y_loc - yerr_loc)

            if ymax is None:
                ymax = ymax_loc
                ymin = ymin_loc
            else:
                ymax = max(ymax, ymax_loc)
                ymin = min(ymin, ymin_loc)

        if line_plot:
            # plot means
            x = time_points + t_add
            for i, cell in enumerate(categories.cells):
                y = line_values[i]

                color = colors[cell]

                if hatch:
                    ls = defaults['linestyle'][i]
                    if color == '1.':
                        color = '0.'
                else:
                    ls = '-'

                if ls == '-':
                    mfc = color
                else:
                    mfc = '1.'

                try:
                    marker = markers[i]
                except:
                    marker = None

        ### NOT same as timeplot() #####  #####  #####  #####  #####  #####
                if self.legend_has_label(cell):
                    label = None
                else:
                    label = cell
                handles = ax.plot(x, y, color=color, linestyle=ls, label=label,
                                  zorder=6, marker=marker, mfc=mfc)
        ### same as timeplot()     #####  #####  #####  #####  #####  #####
                self.add_legend_handles(*handles)

                if spread:
                    if self._x_jitter:
                        x_errbars = x + rel_pos[i]
                    else:
                        x_errbars = x
                    ax.errorbar(x_errbars, y, yerr=yerr[i], fmt=None, zorder=5,
                                ecolor=color, linestyle=ls, label=cell)

        ### NOT same as timeplot() #####  #####  #####  #####  #####  #####
        # heading
        y_unit = (ymax - ymin) / 10
        if y_unit > self._y_unit:
            for forced_y, ln, hd in self._headings:
                if forced_y is not None:
                    hd.set_y(forced_y + self._y_unit / 2)
            self._y_unit = y_unit

        if heading:
            x0 = time_points[ 0] + t_add - tskip / 4
            x1 = time_points[-1] + t_add + tskip / 4
            x = (x0 + x1) / 2
            if headingy is None:
                y = ymax + self._y_unit
                y_text = y + self._y_unit / 2
                if self._heading_y is None:
                    self._heading_y = y
                elif y > self._heading_y:
                    self._heading_y = y
                    for forced_y, ln, hd in self._headings:
                        if forced_y is None:
                            hd.set_y(y_text)
                            ln.set_ydata([y, y])
                        # adjust y
                elif y < self._heading_y:
                    y = self._heading_y
                    y_text = self._heading_y + self._y_unit / 2
            else:
                y = headingy
                y_text = y + self._y_unit / 2

            hd = ax.text(x, y_text, heading, va='bottom', ha='center')
            ln = ax.plot([x0, x1], [y, y], c='k')[0]

            self._headings.append((headingy, ln, hd))
            if not self._ylim:
                ax.set_ylim(top=(y + 3 * y_unit))

        # finalize
        ax.set_xlim(t_min, t_max)
        for t in time_points:
            self._xticks.append(t + t_add)
            self._xticklabels.append('%g' % t)

        if self._ylim:
            ax.set_ylim(self._ylim)

        ax.set_xticks(self._xticks)
        ax.set_xticklabels(self._xticklabels)

        self._show()

    def add_legend(self, fig=False, loc=0, zorder=-1, **kwargs):
        if fig:
            self.figure.figlegend(loc, **kwargs)
        else:
            l = self._ax.legend(loc=loc, **kwargs)
            l.set_zorder(zorder)

        self._show()



def _reg_line(Y, reg):
    coeff = np.hstack([np.ones((len(reg), 1)), reg[:, None]])
    (a, b), residues, rank, s = np.linalg.lstsq(coeff, Y)
    regline_x = np.array([min(reg), max(reg)])
    regline_y = a + b * regline_x
    return regline_x, regline_y


class Correlation(_EelFigure):

    def __init__(self, Y, X, cat=None, sub=None, ds=None, title=None,
                 c=['b', 'r', 'k', 'c', 'm', 'y', 'g'], delim=' ',
                 lloc='lower center', lncol=2, figlegend=True, xlabel=True,
                 ylabel=True, rinxlabel=True, **layout):
        """
        Plot the correlation between two variables.

        Parameters
        ----------
        cat :
            categories
        rinxlabel :
            print the correlation in the xlabel
        """
        _EelFigure.__init__(self, "Correlation", 1, layout, 1, 5,
                            figtitle=title)

        sub = assub(sub, ds)
        Y = asvar(Y, sub, ds)
        X = asvar(X, sub, ds)
        if cat is not None:
            cat = ascategorial(cat, sub, ds)

        # determine labels
        if xlabel is True:
            xlabel = str2tex(X.name)
        if ylabel is True:
            ylabel = str2tex(Y.name)
        if rinxlabel:
            temp = "\n(r={r:.3f}{s}, p={p:.4f}, n={n})"
            if cat is None:
                r, p, n = test._corr(Y, X)
                s = test.star(p)
                xlabel += temp.format(r=r, s=s, p=p, n=n)
            else:
                pass

        # decorate axes
        ax = self._axes[0]
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)

        if cat is None:
            ax.scatter(X.x, Y.x, alpha=.5)
        else:
            labels = []; handles = []
            for color, cell in zip(c, cat.cells):
                idx = (cat == cell)
                Xi = X[idx]
                Yi = Y[idx]
                cell = str2tex(cellname(cell))

                h = ax.scatter(Xi.x, Yi.x, c=color, label=cell, alpha=.5)
                handles.append(h)
                labels.append(cell)

            if figlegend:
                self.figure.legend(handles, labels, lloc, ncol=lncol)
            else:
                ax.legend(lloc, ncol=lncol)

        self._show()


class Regression(_EelFigure):
    def __init__(self, Y, X, cat=None, match=None, sub=None, ds=None,
                 ylabel=True, title=None, alpha=.2, legend=True, delim=' ',
                 c=['#009CFF', '#FF7D26', '#54AF3A', '#FE58C6', '#20F2C3'],
                 **layout):
        """
        Plot the regression of Y on a regressor X.


        parameters
        ----------

        alpha : scalar
            alpha for individual data points (to control visualization of
            overlap)
        legend : bool | str
            applies if cat != None: can be mpl ax.legend() loc kwarg
            http://matplotlib.sourceforge.net/api/axes_api.html#matplotlib.axes.Axes.legend

        """
        sub = assub(sub, ds)
        Y = asvar(Y, sub, ds)
        X = asvar(X, sub, ds)
        if cat is not None:
            cat = ascategorial(cat, sub, ds)
        if match is not None:
            raise NotImplementedError("match kwarg")

        if ylabel is True:
            ylabel = Y.name

        _EelFigure.__init__(self, "Regression", 1, layout, 1, 5, figtitle=title)
        ax = self._axes[0]

        # labels
        if ylabel:
            ax.set_ylabel(ylabel)
        ax.set_xlabel(X.name)
        if title:
            ax.set_title(title, **defaults['title_kwargs'])
        # regplot
        scatter_kwargs = {'s': 100,
                          'alpha': alpha,
                           'marker': 'o',
                           'label': '_nolegend_'}
        if cat is None:
            if type(c) in [list, tuple]:
                color = c[0]
            else:
                color = c
            y = Y.x
            reg = X.x
            ax.scatter(reg, y, edgecolor=color, facecolor=color, **scatter_kwargs)
            x, y = _reg_line(y, reg)
            ax.plot(x, y, c=color)
        else:
            for i, cell in enumerate(cat.cells):
                idx = (cat == cell)
                # scatter
                y = Y.x[idx]
                reg = X.x[idx]
                color = c[i % len(c)]
                ax.scatter(reg, y, edgecolor=color, facecolor=color, **scatter_kwargs)
                # regression line
                x, y = _reg_line(y, reg)
                ax.plot(x, y, c=color, label=cellname(cell))
            if legend == True:
                ax.legend()
            elif legend != False:
                ax.legend(loc=legend)

        self._show()


# MARK: Requirements for Statistical Tests

def _difference(data, names):
    "data condition x subject"
    data_differences = []; diffnames = []; diffnames_2lines = []
    for i, (name1, data1) in enumerate(zip(names, data)):
        for name2, data2 in zip(names[i + 1:], data[i + 1:]):
            data_differences.append(data1 - data2)
            diffnames.append('-'.join([name1, name2]))
            diffnames_2lines.append('-\n'.join([name1, name2]))
    return data_differences, diffnames, diffnames_2lines




def _normality_plot(ax, data, **kwargs):
    """helper fubction for creating normality test figure"""
    n, bins, patches = ax.hist(data, normed=True, **kwargs)
    data = np.ravel(data)

    # normal line
    mu = np.mean(data)
    sigma = np.std(data)
    y = mpl.mlab.normpdf(bins, mu, sigma)
    ax.plot(bins, y, 'r--', linewidth=1)

    # TESTS
    # test Anderson
    A2, thresh, sig = scipy.stats.morestats.anderson(data)
    index = sum(A2 >= thresh)
    if index > 0:
        ax.text(.95, .95, '$^{*}%s$' % str(sig[index - 1] / 100), color='r', size=11,
                horizontalalignment='right',
                verticalalignment='top',
                transform=ax.transAxes,)
    logging.debug(" Anderson: %s, %s, %s" % (A2, thresh, sig))
    # test Lilliefors
    n_test = test.lilliefors(data)
    ax.set_xlabel(r"$D=%.3f$, $p_{est}=%.2f$" % n_test)  # \chi ^{2}
    # make sure ticks display int values
    # ax.yaxis.set_major_formatter(ticker.MaxNLocator(nbins=8, integer=True))
    ticks = ax.get_yticks()
    ticks_int = [int(l) for l in ticks]
    ax.set_yticks(ticks_int)
    ax.set_yticklabels(ticks_int)


class Histogram(_EelFigure):
    def __init__(self, Y, X=None, match=None, sub=None, ds=None, pooled=True,
                title=True, ylabel=True, **layout):
        """Make histogram plots and test normality.

        Parameters
        ----------
        pooled : bool
            Add one plot with all values/differences pooled.
        title : None | str
            Figure title.
        titlekwargs : dict
            Forwarded to :py:func:`pyplot.suptitle`.
        """
        ct = Celltable(Y, X, match=match, sub=sub, ds=ds, coercion=asvar)

        # ylabel
        if ylabel is True:
            ylabel = str2tex(getattr(ct.Y, 'name', False))

        # layout
        if X is None:
            nax = 1
            if title is True:
                title = "Test for Normality"
        elif ct.all_within:
            nax = None
            if title is True:
                title = "Tests for Normality of the Differences"
        else:
            nax = len(ct.cells)
            if title is True:
                title = "Tests for Normality"

        _EelFigure.__init__(self, "Histogram", nax, layout, 1, 4,
                            figtitle=title)

        if X is None:
            ax = self._axes[0]
            _normality_plot(ax, ct.Y.x)
        elif ct.all_within:
            # temporary:
            data = ct.get_data()
            names = ct.cellnames()

#             plt.subplots_adjust(hspace=.5)

            n_cmp = len(ct.cells) - 1
            pooled = []
            # i: row
            # j: -column
            for i in range(0, n_cmp + 1):
                for j in range(i + 1, n_cmp + 1):
                    difference = data[i] - data[j]
                    pooled.append(scipy.stats.zscore(difference))  # z transform?? (scipy.stats.zs())
                    ax_i = n_cmp * i + (n_cmp + 1 - j)
                    ax = self.figure.add_subplot(n_cmp, n_cmp, ax_i)
                    _normality_plot(ax, difference)
                    if i == 0:
                        ax.set_title(names[j], size=12)
                    if j == n_cmp:
                        ax.set_ylabel(names[i], size=12)
            # pooled diffs
            if len(names) > 2:
                ax = self.figure.add_subplot(n_cmp, n_cmp, n_cmp ** 2)
                _normality_plot(ax, pooled, facecolor='g')
                ax.set_title("Pooled Differences (n=%s)" % len(pooled),
                             weight='bold')
                self.figure.text(.99, .01, "$^{*}$ Anderson and Darling test "
                                 "thresholded at $[ .15,   .10,    .05,    "
                                 ".025,   .01 ]$.", color='r',
                                 verticalalignment='bottom',
                                 horizontalalignment='right')
        else:  # independent measures
#             plt.subplots_adjust(hspace=.5, left=.1, right=.9, bottom=.1, top=.8)
            for cell, ax in izip(ct.cells, self._axes):
                ax.set_title(cellname(cell))
                _normality_plot(ax, ct.data[cell])


def boxcox_explore(Y, params=[-1, -.5, 0, .5, 1], crange=False, ax=None, box=True):
    """
    Plot the distribution resulting from a Box-Cox transformation with
    different parameters.


    Parameters
    ----------

    Y : array-like
        array containing data whose distribution is to be examined
    crange :
        correct range of transformed data
    ax :
        ax to plot to (``None`` creates a new figure)
    box :
        use Box-Cox family

    """
    if hasattr(Y, 'x'):
        Y = Y.x
    else:
        Y = np.ravel(Y)

    if np.any(Y == 0):
        raise ValueError("data contains 0")

    y = []
    for p in params:
        if p == 0:
            if box:
                xi = np.log(Y)
            else:
                xi = np.log10(Y)
                # xi = np.log1p(x)
        else:
            if box:
                xi = (Y ** p - 1) / p
            else:
                xi = Y ** p
        if crange:
            xi -= min(xi)
            xi /= max(xi)
        y.append(xi)

    if not ax:
        plt.figure()
        ax = plt.subplot(111)

    ax.boxplot(y)
    ax.set_xticks(range(1, 1 + len(params)))
    ax.set_xticklabels(params)
    ax.set_xlabel("p")
    if crange:
        ax.set_ylabel("Value (Range Corrected)")


# backwards comp
barplot = Barplot
boxplot = Boxplot
corrplot = Correlation
histogram = Histogram
multitimeplot = MultiTimeplot
regplot = Regression
timeplot = Timeplot



'''
Plot uniform time-series of one variable.
'''
# author: Christian Brodbeck
from __future__ import division

import numpy as np
import scipy.stats
import matplotlib.cm as _cm

from .._data_obj import ascategorial, asndvar, cellname, Celltable
from . import _base


class UTSStat(_base._EelFigure):
    "Plots statistics for a one-dimensional NDVar"
    def __init__(self, Y='Y', X=None, Xax=None, match=None, sub=None, ds=None,
                 main=np.mean, dev=scipy.stats.sem, legend='upper right',
                 title=None, axtitle='{name}', xlabel=True, ylabel=True,
                 invy=False, bottom=None, top=None, hline=None, xdim='time',
                 xlim=None, color='b', colors='jet', clusters=None,
                 frame=True, **layout):
        """
    Plot statistics for a one-dimensional NDVar

    Parameters
    ----------
    Y : 1d-NDVar
        Dependent variable (one-dimensional NDVar).
    X : categorial or None
        Model: specification of conditions which should be plotted separately.
    Xax : None | categorial
        Make separate axes for each category in this categorial model.
    match : Factor
        Identifier for repeated measures data.
    sub : None | index array
        Only use a subset of the data provided.
    ds : None | Dataset
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    main : func | None
        Measure for the central tendency (function that takes an ``axis``
        argument). The default is numpy.mean.
    dev : func | 'all' | float
        Measure for spread / deviation from the central tendency. Either a
        function that takes an ``axis`` argument, 'all' to plot all traces, or
        a float to plot all traces with a certain alpha value. The default is
        numpy.stats.sem which plots the standard error of the mean.
    legend : str | None
        matplotlib figure legend location argument
    title : str | None
        Figure title; '{name}' will be formatted to ``Y.name``
    axtitle : str | None
        Axes title. '{name}' is formatted to the category name.
    xlabel, ylabel : True |str | None
        X- and y axis label. If True the labels will be inferred from the data.
    invy : bool
        invert the y axis
    bottom, top | None | scalar
        Set an absolute range for the plot's y axis.
    hline : None | scalar | (value, kwarg-dict) tuple
        Add a horizontal line to each plot. If provided as a tuple, the second
        element can include any keyword arguments that should be submitted to
        the call to matplotlib axhline call.
    xdim : str
        dimension for the x-axis (default is 'time')
    xlim : None | (scalar, scalar)
        Tuple of xmin and xmax to set the initial x-axis limits.
    color : matplotlib color
        Color if just a single category of data is plotted.
    colors : str | list | dict
        Colors for the plots if multiple categories of data are plotted.
        **str**: A colormap name; Cells of X are mapped onto the colormap in
        regular intervals.
        **list**: A list of colors in the same sequence as X.cells.
        **dict**: A dictionary mapping each cell in X to a color.
        Colors are specified as `matplotlib compatible color arguments
        <http://matplotlib.org/api/colors_api.html>`_.
    clusters : None | Dataset
        Clusters to add to the plots. The clusters should be provided as
        Dataset, as stored in test results' :py:attr:`.clusters`.
    frame : bool
        Draw a frame containing the figure from the top and the right (default
        ``True``).
        """
        if Xax is None:
            nax = 1
            ct = Celltable(Y, X, sub=sub, match=match, ds=ds, coercion=asndvar)
            if X is None:
                cells = None
            else:
                cells = ct.X.cells
        else:
            ct = Celltable(Y, Xax, sub=sub, ds=ds, coercion=asndvar)
            if X is None:
                cells = None
                X_ = None
            else:
                Xct = Celltable(X, Xax, sub=sub, ds=ds, coercion=ascategorial)
                cells = Xct.Y.cells
            if match is not None:
                matchct = Celltable(match, Xax, sub=sub, ds=ds,
                                    coercion=ascategorial)
            nax = len(ct.cells)

        # assemble colors
        if cells is None:
            colors = {None: color}
        else:
            if isinstance(colors, (list, tuple)):
                if len(colors) < len(cells):
                    err = ("The `colors` argument %s does not supply enough "
                           "colors (%i) for %i "
                           "cells." % (str(colors), len(colors), len(cells)))
                    raise ValueError(err)
                colors = dict(zip(cells, colors))
            elif isinstance(colors, dict):
                for cell in cells:
                    if cell not in colors:
                        raise KeyError("%s not in colors" % repr(cell))
            else:
                cm = _cm.get_cmap(colors)
                N = len(cells)
                colors = {cell: cm(i / N) for i, cell in enumerate(cells)}

        kwargs = dict(dev=dev, main=main, ylabel=ylabel, xdim=xdim,
                      invy=invy, bottom=bottom, top=top, hline=hline,
                      xlabel=xlabel, xlim=xlim, frame=frame)

        if title is not None and '{name}' in title:
            title = title.format(name=ct.Y.name)
        super(UTSStat, self).__init__("UTSStat Plot", nax, layout,
                                      figtitle=title)

        self._plots = []
        self._legend_handles = {}
        if Xax is None:
            ax = self._axes[0]
            if axtitle and '{name}' in axtitle:
                title_ = axtitle.format(name=ct.Y.name)
            else:
                title_ = axtitle
            p = _ax_stat(ax, ct, colors, title=title_, **kwargs)
            self._plots.append(p)
            self._legend_handles.update(p.legend_handles)
            if len(ct) < 2:
                legend = False
        else:
            for i, ax, cell in zip(xrange(nax), self._axes, ct.cells):
                kwargs['xlabel'] = xlabel if i == len(ct) - 1 else False
                if X is not None:
                    X_ = Xct.data[cell]
                if match is not None:
                    match = matchct.data[cell]
                cct = Celltable(ct.data[cell], X_, match=match,
                                coercion=asndvar)
                title_ = axtitle.format(name=cellname(cell))
                p = _ax_stat(ax, cct, colors, title=title_, **kwargs)
                self._plots.append(p)
                self._legend_handles.update(p.legend_handles)

        self.plot_legend(legend)

        # prepare cluster plots
        self._clusters = [None] * nax
        self._clusters_all_same = True
        if clusters is not None:
            self.set_clusters(clusters)

        self._show()

    def _fill_toolbar(self, tb):
        import wx

        btn = self._cluster_btn = wx.Button(tb, wx.ID_ABOUT, "Clusters")
        btn.Enable(False)
        tb.AddControl(btn)
        btn.Bind(wx.EVT_BUTTON, self._OnShowClusterInfo)

    def _OnShowClusterInfo(self, event):
        from ..._wxutils import show_text_dialog
        if self._clusters_all_same:
            info = str(self._clusters)
        else:
            info = []
            for i, clusters in enumerate(self._clusters):
                if clusters is None:
                    continue
                title = "Axes %i" % i
                info.append(title)
                info.append('\n')
                info.append('-' * len(title))
                info.append(str(clusters))
            info = '\n'.join(info)

        show_text_dialog(self._frame, info, "Clusters")

    def plot_legend(self, loc='fig', figsize=(2, 2)):
        """Plots (or removes) the legend from the figure.

        Parameters
        ----------
        loc : False | 'fig' | str | int
            Where to plot the legend (see Notes; default 'fig').
        figsize : tuple of 2 int
            Figure size if plotting to figure (i.e., loc == 'fig').

        Returns
        -------
        legend_figure : None | legend
            If loc=='fig' the Figure, otherwise None.

        Notes
        -----
        legend content can be modified through the figure's
        ``legend_handles`` and ``legend_labels`` attributes.

        Possible values for the ``loc`` argument:

        ``False``:
            Make the current legend invisible
        'fig':
            Plot the legend in a new figure
        str | int:
            Matplotlib position argument: plot the legend on the figure


        Matplotlib Position Arguments:

         - 'upper right'  : 1,
         - 'upper left'   : 2,
         - 'lower left'   : 3,
         - 'lower right'  : 4,
         - 'right'        : 5,
         - 'center left'  : 6,
         - 'center right' : 7,
         - 'lower center' : 8,
         - 'upper center' : 9,
         - 'center'       : 10,
        """
        if loc and len(self._legend_handles) > 1:
            cells = sorted(self._legend_handles)
            labels = [cellname(cell) for cell in cells]
            handles = [self._legend_handles[cell] for cell in cells]
            if loc == 'fig':
                return _base.legend(handles, labels, figsize=figsize)
            else:
                # take care of old legend; remove() not implemented as of mpl 1.3
                if hasattr(self, 'legend'):
                    self.legend.set_visible(False)
                self.legend = self.figure.legend(handles, labels, loc=loc)
                self.draw()
        else:
            if hasattr(self, 'legend'):
                self.legend.set_visible(False)
                del self.legend
                self.draw()

    def set_clusters(self, clusters, pmax=0.05, ptrend=0.1, color='.7', ax=None):
        """Add clusters from a cluster test to the plot (as shaded area).

        Parameters
        ----------
        clusters : None | Dataset
            The clusters, as stored in test results' :py:attr:`.clusters`.
            Use ``None`` to remove the clusters plotted on a given axis.
        pmax : scalar
            Maximum p-value of clusters to plot as solid.
        ptrend : scalar
            Maximum p-value of clusters to plot as trend.
        color : matplotlib color
            Color for the clusters.
        ax : None | int
            Index of the axes to which the clusters are to be added. If None,
            add the clusters to all axes.
        """
        nax = len(self._axes)
        if ax is None:
            self._clusters_all_same = True
            self._clusters = clusters
            axes = xrange(nax)
        else:
            if self._clusters_all_same:
                self._clusters = [self._clusters] * nax
                self._clusters_all_same = False
            self._clusters[ax] = clusters
            axes = [ax]

        # update plots
        for ax in axes:
            p = self._plots[ax].cluster_plt
            p.set_clusters(clusters, False)
            p.set_color(color, False)
            p.set_pmax(pmax, ptrend)
        self.draw()

        # update GUI
        if hasattr(self, '_cluster_btn'):
            if clusters is not None:
                self._cluster_btn.Enable(True)
            elif self._clusters_all_same or all(c is None for c in self._clusters):
                self._cluster_btn.Enable(False)

    def set_xlim(self, xmin, xmax):
        """Adjust the x-axis limits on all axes
        """
        for ax in self._axes:
            ax.set_xlim(xmin, xmax)
        self.draw()

    def set_ylim(self, bottom=None, top=None):
        """
        Adjust the y-axis limits on all axes (see matplotlib's
        :py:meth:`axes.set_ylim`)

        """
        for ax in self._axes:
            ax.set_ylim(bottom, top)

        self.draw()



class UTS(_base._EelFigure):
    "Value by time plot for UTS data."
    def __init__(self, epochs, Xax=None, title=None, axtitle='{name}', ds=None,
                 **layout):
        """Value by time plot for UTS data

        Parameters
        ----------
        epochs : epochs
            Uts data epochs to plot.
        Xax : None | categorial
            Make separate axes for each category in this categorial model.
        title : None | str
            Figure title.
        axtitle : None | str
            Axes title. '{name}' is formatted to the category name.
        ds : None | Dataset
            If a Dataset is specified, all data-objects can be specified as
            names of Dataset variables.
        """
        epochs = self.epochs = _base.unpack_epochs_arg(epochs, 1, Xax, ds)
        super(UTS, self).__init__("UTS", len(epochs), layout, 1.5, 2,
                                  figtitle=title)

        for ax, epoch in zip(self._axes, epochs):
            _ax_uts(ax, epoch, title=axtitle)

        self._show()


class _ax_stat:
    def __init__(self, ax, ct, colors, dev=scipy.stats.sem,
                 main=np.mean, title=True, ylabel=True, xdim='time', xlim=None,
                 xlabel=True, invy=False, bottom=None, top=None, hline=None,
                 clusters=None, pmax=0.05, ptrend=0.1, frame=True):
        ax.x_fmt = "t = %.3f s"

        # stat plots
        self.stat_plots = []
        self.legend_handles = {}

        x = ct.Y.get_dim(xdim)
        for cell in ct.cells:
            cell_label = cellname(cell)
            c = colors[cell]
            ndvar = ct.data[cell]
            y = ndvar.get_data(('case', xdim))
            plt = _plt_stat(ax, x, y, main, dev, label=cell_label, color=c)
            self.stat_plots.append(plt)
            if plt.main is not None:
                self.legend_handles[cell] = plt.main[0]

        # hline
        if hline is not None:
            if isinstance(hline, tuple):
                if len(hline) != 2:
                    raise ValueError("hline must be None, scalar or length 2 tuple")
                hline, hline_kw = hline
                hline_kw = dict(hline_kw)
            else:
                hline_kw = {'color': 'k'}

            hline = float(hline)
            ax.axhline(hline, **hline_kw)

        # cluster plot
        self.cluster_plt = _plt_uts_clusters(ax, clusters, pmax, ptrend)

        # title
        if title:
            if title is True:
                title = ct.Y.name
            ax.set_title(title)

        # format x axis
        ax.set_xlim(x[0], x[-1])
        ax.x_fmt = "t = %.3f s"
        ticks = ax.xaxis.get_ticklocs()
        ticklabels = _base._ticklabels(ticks, 'time')
        ax.xaxis.set_ticklabels(ticklabels)

        _base.set_xlabel(ax, xdim, xlabel)

        if ylabel is True:
            ylabel = ct.Y.info.get('unit', None)
        if ylabel:
            ax.set_ylabel(ylabel)

        if invy:
            y0, y1 = ax.get_ylim()
            bottom = bottom if (bottom is not None) else y1
            top = top if (top is not None) else y0
        if (bottom is not None) or (top is not None):
            ax.set_ylim(bottom, top)

        if xlim is None:
            xlim = (min(x), max(x))
        xmin, xmax = xlim
        ax.set_xlim(xmin, xmax)

        if not frame:
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ax.yaxis.set_ticks_position('left')
            ax.xaxis.set_ticks_position('bottom')

        # store attributes
        self.ax = ax
        self.title = title


class UTSClusters(_base._EelFigure):
    "Plotting of ANOVA permutation cluster test results"
    def __init__(self, res, pmax=0.05, ptrend=0.1, title=None,
                 axtitle='{name}', cm='jet', overlay=False, **layout):
        """
        Plotting of permutation cluster test results

        Parameters
        ----------
        res : testnd.anova
            ANOVA with permutation cluster test result object.
        pmax : scalar
            Maximum p-value of clusters to plot as solid.
        ptrend : scalar
            Maximum p-value of clusters to plot as trend.
        title : str
            Figure title.
        axtitle : None | str
            Axes title pattern. '{name}' is formatted to the effect name.
        cm : str
            Colormap to use for coloring different effects.
        overlay : bool
            Plot epochs (time course for different effects) on top of each
            other (as opposed to on separate axes).
        """
        clusters_ = res.clusters

        epochs = self.epochs = _base.unpack_epochs_arg(res, 1)
        cm = _cm.get_cmap(cm)

        # create figure
        N = len(epochs)
        nax = 1 if overlay else N
        super(UTSClusters, self).__init__("UTSClusters", nax, layout,
                                          figtitle=title)

        ylabel = True
        self._caxes = []
        if overlay:
            ax = self._axes[0]
            axtitle = None
        for i, layers in enumerate(epochs):
            if not overlay:
                ax = self._axes[i]

            color = cm(i / N)
            stat = layers[0]

            # ax clusters
            if clusters_:
                if 'effect' in clusters_:
                    cs = clusters_.sub('effect == %r' % layers[0].name)
                else:
                    cs = clusters_
            else:
                cs = None

            cax = _ax_uts_clusters(ax, stat, cs, color, pmax, ptrend, 'time',
                                   axtitle, ylabel)
            self._caxes.append(cax)
            ylabel = None

        self.clusters = clusters_
        self._show()

    def _fill_toolbar(self, tb):
        import wx

        btn = wx.Button(tb, wx.ID_ABOUT, "Clusters")
        tb.AddControl(btn)
        btn.Bind(wx.EVT_BUTTON, self._OnShowClusterInfo)

    def _OnShowClusterInfo(self, event):
        from ..._wxutils import show_text_dialog
        info = str(self.clusters)
        show_text_dialog(self._frame, info, "Clusters")

    def set_pmax(self, pmax=0.05, ptrend=0.1):
        "set the threshold p-value for clusters to be displayed"
        for cax in self._caxes:
            cax.set_pmax(pmax, ptrend)
        self.draw()


def _ax_uts(ax, layers, title=False, bottom=None, top=None, invy=False,
            xlabel=True, ylabel=True, color=None, xdim='time'):
    contours = {}
    overlay = False
    for l in layers:
        args = _base.find_uts_args(l, overlay, color)
        overlay = True
        if args is None:
            continue

        _plt_uts(ax, l, xdim=xdim, **args)
        contours = l.info.get('contours', None)
        if contours:
            for v, color in contours.iteritems():
                if v in contours:
                    continue
                contours[v] = ax.axhline(v, color=color)

    l0 = layers[0]
    x = l0.get_dim(xdim)
    ax.set_xlim(x[0], x[-1])
    ax.x_fmt = "t = %.3f s"

    ticks = ax.xaxis.get_ticklocs()
    ticklabels = _base._ticklabels(ticks, 'time')
    ax.xaxis.set_ticklabels(ticklabels)

    if title:
        if 'name' in title:
            title = title.format(name=l0.name)
        ax.set_title(title)

    _base.set_xlabel(ax, xdim, xlabel)

    if ylabel is True:
        ylabel = l.info.get('unit', None)
    if ylabel:
        ax.set_ylabel(ylabel)

    if invy:
        y0, y1 = ax.get_ylim()
        bottom = bottom if (bottom is not None) else y1
        top = top if (top is not None) else y0
    if (bottom is not None) or (top is not None):
        ax.set_ylim(bottom, top)


def _plt_uts(ax, ndvar, color=None, xdim='time', kwargs={}):
    y = ndvar.get_data((xdim,))
    x = ndvar.get_dim(xdim).x
    if color is not None:
        kwargs['color'] = color
    ax.plot(x, y, **kwargs)

    for y, kwa in _base.find_uts_hlines(ndvar):
        if color is not None:
            kwa['color'] = color
        ax.axhline(y, **kwa)


class _ax_uts_clusters:
    def __init__(self, ax, Y, clusters, color=None, pmax=0.05, ptrend=0.1,
                 xdim='time', title=None, xlabel=True, ylabel=True):
        uts_args = _base.find_uts_args(Y, False, color)
        self._bottom, self._top = _base.find_vlim_args(Y)

        if title:
            if '{name}' in title:
                title = title.format(name=Y.name)
            ax.set_title(title)

        _plt_uts(ax, Y, xdim=xdim, **uts_args)

        if ylabel is True:
            ylabel = Y.info.get('meas', _base.default_meas)
        if ylabel:
            ax.set_ylabel(ylabel)

        xlabel = _base._axlabel(xdim, xlabel)
        if xlabel:
            ax.set_xlabel(xlabel)
        if np.any(Y.x < 0) and np.any(Y.x > 0):
            ax.axhline(0, color='k')


        # pmap
        self.cluster_plt = _plt_uts_clusters(ax, clusters, pmax, ptrend, color)

        # save ax attr
        self.ax = ax
        x = Y.get_dim(xdim).x
        self.xlim = (x[0], x[-1])

        ax.set_xlim(*self.xlim)
        ax.set_ylim(bottom=self._bottom, top=self._top)

    def set_clusters(self, clusters):
        self.cluster_plt.set_clusters(clusters)

    def set_pmax(self, pmax=0.05, ptrend=0.1):
        self.cluster_plt.set_pmax(pmax, ptrend)


class _plt_uts_clusters:
    def __init__(self, ax, clusters, pmax, ptrend, color=None, hatch='/'):
        """
        clusters : Dataset
            Dataset with entries for 'tstart', 'tstop' and 'p'.
        """
        self.pmax = pmax
        self.ptrend = ptrend
        self.h = []
        self.ax = ax
        self.clusters = clusters
        self.color = color
        self.hatch = hatch
        self.update()

    def set_clusters(self, clusters, update=True):
        self.clusters = clusters
        if update:
            self.update()

    def set_color(self, color, update=True):
        self.color = color
        if update:
            self.update()

    def set_pmax(self, pmax, ptrend, update=True):
        self.pmax = pmax
        self.ptrend = ptrend
        if update:
            self.update()

    def update(self):
        h = self.h
        while len(h):
            h.pop().remove()

        clusters = self.clusters
        if clusters is None:
            return

        p_include = self.ptrend or self.pmax
        for cluster in clusters.itercases():
            if 'p' in cluster:
                p = cluster['p']
                if p > p_include:
                    continue
                alpha = 0.5 if p < self.pmax else 0.2
            else:
                alpha = 0.5

            x0 = cluster['tstart']
            x1 = cluster['tstop']
            h = self.ax.axvspan(x0, x1, color=self.color,  # , hatch=self.hatch,
                                fill=True, alpha=alpha, zorder=-1)
            self.h.append(h)


class _plt_stat(object):
    def __init__(self, ax, x, y, main, dev, label=None, color=None, **kwargs):
        if color:
            kwargs['color'] = color

        main_kwargs = kwargs.copy()
        dev_kwargs = kwargs.copy()
        if label:
            main_kwargs['label'] = label

        if isinstance(dev, basestring):
            if dev == 'all':
                dev_kwargs['alpha'] = .3
            else:
                err = ("The only possible str value for dev is 'all'; got "
                       "%r." % dev)
                raise ValueError(err)
        elif np.isscalar(dev):
            dev_kwargs['alpha'] = dev
            dev = 'all'
        else:
            dev_kwargs['alpha'] = .3

        if dev == 'all':
            if 'linewidth' in kwargs:
                main_kwargs['linewidth'] = kwargs['linewidth'] * 2
            elif 'lw' in kwargs:
                main_kwargs['lw'] = kwargs['lw'] * 2
            else:
                main_kwargs['lw'] = 2

        # plot main
        if hasattr(main, '__call__'):
            y_ct = main(y, axis=0)
            self.main = ax.plot(x, y_ct, zorder=5, **main_kwargs)
        elif dev == 'all':
            self.main = None
        else:
            raise ValueError("Invalid argument: main=%r" % main)

        # plot dev
        if hasattr(dev, '__call__'):
            ydev = dev(y, axis=0)
            self.dev = ax.fill_between(x, y_ct - ydev, y_ct + ydev, zorder=0, **dev_kwargs)
        elif dev == 'all':
            self.dev = ax.plot(x, y.T, **dev_kwargs)
        else:
            self.dev = None



'''
Created on Apr 22, 2013

@author: Christian M Brodbeck
'''
import numpy as np
from scipy.io import savemat
import wx



def draw_text(text, face='Scheherazade', size=42, spo2=False, w=None, h=None, color=None):
    """
    Use wxPython's text rendering engine to convert unicode text to an image
    in the form of a numpy array.

    Parameters
    ----------
    text : unicode
        The text.
    face : str
        The font name.
    size : int
        Font size.
    spo2 : bool
        If the shape is determined automatically (with shape=None), use a
        Square with a side length that its a Power Of 2.
    w, h : None | int
        Specify the desired width and/or height for the output array.
        If None, the shape is determined based on the text size.
    color : None | array, len = 3
        Color of the text. If None (monochrome), an array of shape (h, w) is
        returned. If an array of three numbers, they are interpreted as RGB
        values and an array of shape (h, w, 3) is returned.
        .
    """
    if (w is not None or h is not None) and spo2:
        err = ("spo2 can not be true when providing custom width and/ir "
               "height values.")
        raise TypeError(err)

    font = wx.Font(size, family=wx.FONTFAMILY_UNKNOWN, style=wx.FONTSTYLE_NORMAL,
                   weight=wx.FONTWEIGHT_NORMAL, face=face)

    dc = wx.MemoryDC()
    dc.SetFont(font)
    res = 512
    dc.SelectObject(wx.EmptyBitmap(res, res))

    tw, th = dc.GetTextExtent(text)
    if w is None:
        w = tw
    if h is None:
        h = th

    if spo2:
        x = 2
        while (x < w) or (x < h):
            x *= 2
        w, h = x, x
    elif tw > w or th > h:
        err = ("Text dimensions (w=%i, h=%i) exceed image resolution "
               "(w=%i, h=%i)." % (tw, th, w, h))
        raise ValueError(err)

    bmp = wx.EmptyBitmap(w, h)
    dc.SelectObject(bmp)
    dc.Clear()

    dc.DrawText(text, (w - tw) / 2, (h - th) / 2)

    im = bmp.ConvertToImage()
    a = np.fromstring(im.GetData(), dtype='uint8').reshape((h, w, 3))
    if color is None:
        return 255 - a[:, :, 0]
    else:
        return a * np.reshape(color, (1, 1, 3))


def draw_paragraph(lines, face='Scheherazade', size=42, align='center'):
    """
    Parameters
    ----------
    lines : iterator
        Iterator over unicode strings, each of which will be added as a
        separate line.
    face : str
        The font name.
    size : int
        Font size.
    align : 'left' | 'center' | 'right'
        How to align lines
    """
    ims = [draw_text(line, face=face, size=size, spo2=False) for line in lines]
    shape = np.array([im.shape for im in ims])
    h = np.sum(shape[:, 0])
    w = np.max(shape[:, 1])
    out = np.zeros((h, w), dtype='uint8')
    y = 0
    for im in ims:
        im_h, im_w = im.shape
        y0 = y
        y += im_h
        if align == 'left':
            x = 0
            x1 = im_w
        elif align == 'center':
            x = (w - im_w) // 2
            x1 = x + im_w
        elif align == 'right':
            x = w - im_w
            x1 = w
        else:
            raise ValueError("align=%r" % align)

        out[y0:y, x:x1] = im

    return out


class UnicodeImages(object):
    """
    Build a dictionary with image arrays depicting unicode strings, ensuring
    the same string is not stored twice.
    """
    def __init__(self, face='Scheherazade', size=42, spo2=False):
        """
        Parameters
        ----------
        draw_text parameters
        """
        self.arrays = {}
        self._codes = {}
        self._face = face
        self._size = size
        self._spo2 = spo2

    def code(self, text):
        """
        Retrieve the code for the text.

        Parameters
        ----------
        text : unicode
            The unicode string to be converted to an image.

        Returns
        -------
        code : int
            The code assigned to the unicode object.
        """
        if text in self._codes:
            code = self._codes[text]
        else:
            im = draw_text(text, face=self._face, size=self._size,
                           spo2=self._spo2)
            code = len(self._codes)
            str_code = 'i%i' % code
            self.arrays[str_code] = im
            self._codes[text] = code

        return code

    def save(self, fname):
        "Save as matlab *.m file."
        savemat(fname, self.arrays, do_compression=True)



import webbrowser

import wx

from .._wxutils import ID, logger
from .about import AboutFrame


class App(wx.App):
    def OnInit(self):
        self.SetExitOnFrameDelete(False)
        self.SetAppName("Eelbrain")

        # File Menu
        m = file_menu = wx.Menu()
        m.Append(wx.ID_OPEN, '&Open... \tCtrl+O')
        m.AppendSeparator()
        m.Append(wx.ID_CLOSE, '&Close Window \tCtrl+W')
        m.Append(wx.ID_SAVE, "Save \tCtrl+S")
        m.Append(wx.ID_SAVEAS, "Save As... \tCtrl+Shift+S")

        # Edit Menu
        m = edit_menu = wx.Menu()
        m.Append(ID.UNDO, '&Undo \tCtrl+Z')
        m.Append(ID.REDO, '&Redo \tCtrl+Shift+Z')
        m.AppendSeparator()
        m.Append(wx.ID_CUT, 'Cut \tCtrl+X')
        m.Append(wx.ID_COPY, 'Copy \tCtrl+C')
        m.Append(wx.ID_PASTE, 'Paste \tCtrl+V')
        m.AppendSeparator()
        m.Append(wx.ID_CLEAR, 'Cle&ar')

        # View Menu
        m = view_menu = wx.Menu()
        m.Append(ID.SET_VLIM, "Set Y-Axis Limit... \tCtrl+l", "Change the Y-"
                 "axis limit in epoch plots")
        m.Append(ID.SET_LAYOUT, "&Set Layout... \tCtrl+Shift+l", "Change the "
                 "page layout")
        m.AppendCheckItem(ID.PLOT_RANGE, "&Plot Data Range \tCtrl+r", "Plot "
                          "data range instead of individual sensor traces")

        # Go Menu
        m = go_menu = wx.Menu()
        m.Append(wx.ID_FORWARD, '&Forward \tCtrl+]', 'Go One Page Forward')
        m.Append(wx.ID_BACKWARD, '&Back \tCtrl+[', 'Go One Page Back')

        # Help Menu
        m = help_menu = wx.Menu()
        m.Append(ID.HELP_EELBRAIN, 'Eelbrain Help')
        m.Append(ID.HELP_PYTHON, "Python Help")
        m.AppendSeparator()
        m.Append(wx.ID_ABOUT, '&About Eelbrain')

        # Menu Bar
        menu_bar = wx.MenuBar()
        menu_bar.Append(file_menu, "File")
        menu_bar.Append(edit_menu, "Edit")
        menu_bar.Append(view_menu, "View")
        menu_bar.Append(go_menu, "Go")
        menu_bar.Append(help_menu, self.GetMacHelpMenuTitleName())
        wx.MenuBar.MacSetCommonMenuBar(menu_bar)
        self.menubar = menu_bar

        # Bind Menu Commands
        self.Bind(wx.EVT_MENU, self.OnAbout, id=wx.ID_ABOUT)
        self.Bind(wx.EVT_MENU, self.OnOpen, id=wx.ID_OPEN)
        self.Bind(wx.EVT_MENU, self.OnClear, id=wx.ID_CLEAR)
        self.Bind(wx.EVT_MENU, self.OnCloseWindow, id=wx.ID_CLOSE)
        self.Bind(wx.EVT_MENU, self.OnCopy, id=wx.ID_COPY)
        self.Bind(wx.EVT_MENU, self.OnCut, id=wx.ID_CUT)
        self.Bind(wx.EVT_MENU, self.OnOnlineHelp, id=ID.HELP_EELBRAIN)
        self.Bind(wx.EVT_MENU, self.OnOnlineHelp, id=ID.HELP_PYTHON)
        self.Bind(wx.EVT_MENU, self.OnPaste, id=wx.ID_PASTE)
        self.Bind(wx.EVT_MENU, self.OnRedo, id=ID.REDO)
        self.Bind(wx.EVT_MENU, self.OnSave, id=wx.ID_SAVE)
        self.Bind(wx.EVT_MENU, self.OnSaveAs, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_MENU, self.OnSetLayout, id=ID.SET_LAYOUT)
        self.Bind(wx.EVT_MENU, self.OnSetVLim, id=ID.SET_VLIM)
        self.Bind(wx.EVT_MENU, self.OnTogglePlotRange, id=ID.PLOT_RANGE)
        self.Bind(wx.EVT_MENU, self.OnUndo, id=ID.UNDO)
        if wx.__version__ < '3':
            self.Bind(wx.EVT_MENU, self.LegacyOnQuit, id=wx.ID_EXIT)

        # bind update UI
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIBackward, id=wx.ID_BACKWARD)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIClear, id=wx.ID_CLEAR)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIClose, id=wx.ID_CLOSE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUICopy, id=wx.ID_COPY)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUICut, id=wx.ID_CUT)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIForward, id=wx.ID_FORWARD)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIOpen, id=wx.ID_OPEN)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIPaste, id=wx.ID_PASTE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIPlotRange, id=ID.PLOT_RANGE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIRedo, id=ID.REDO)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISave, id=wx.ID_SAVE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISaveAs, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISetLayout, id=ID.SET_LAYOUT)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISetVLim, id=ID.SET_VLIM)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIUndo, id=ID.UNDO)

        return True

    def _get_active_frame(self):
        win = wx.Window.FindFocus()
        return wx.GetTopLevelParent(win)
        for w in  wx.GetTopLevelWindows():
            if hasattr(w, 'IsActive') and w.IsActive():
                return w
        return wx.GetActiveWindow()

    def OnAbout(self, event):
        if hasattr(self, '_about_frame') and hasattr(self._about_frame, 'Raise'):
            self._about_frame.Raise()
        else:
            self._about_frame = AboutFrame(None)
            self._about_frame.Show()
#             frame.SetFocus()

    def OnClear(self, event):
        frame = self._get_active_frame()
        frame.OnClear(event)

    def OnCloseWindow(self, event):
        frame = self._get_active_frame()
        logger.debug("Close %r" % frame)
        if frame is not None:
            frame.Close()

    def OnCopy(self, event):
        win = wx.Window.FindFocus()
        logger.debug("Copy %r" % win)
        win.Copy()
#
    def OnCut(self, event):
        win = wx.Window.FindFocus()
        logger.debug("Cut %r" % win)
        win.Cut()

    def OnOnlineHelp(self, event):
        "Called from the Help menu to open external resources"
        Id = event.GetId()
        if Id == ID.HELP_EELBRAIN:
            webbrowser.open("https://pythonhosted.org/eelbrain/")
        elif Id == ID.HELP_PYTHON:
            webbrowser.open("http://docs.python.org/2.7/")
        else:
            raise RuntimeError("Invalid help ID")

    def OnOpen(self, event):
        frame = self._get_active_frame()
        frame.OnOpen(event)

    def OnPaste(self, event):
        win = wx.Window.FindFocus()
        logger.debug("Paste %r" % win)
        win.Paste()

    def OnRedo(self, event):
        frame = self._get_active_frame()
        frame.OnRedo(event)

    def OnSave(self, event):
        frame = self._get_active_frame()
        frame.OnSave(event)

    def OnSaveAs(self, event):
        frame = self._get_active_frame()
        frame.OnSaveAs(event)

    def OnSetVLim(self, event):
        frame = self._get_active_frame()
        frame.OnSetVLim(event)

    def OnSetLayout(self, event):
        frame = self._get_active_frame()
        frame.OnSetLayout(event)

    def OnTogglePlotRange(self, event):
        frame = self._get_active_frame()
        frame.OnTogglePlotRange(event)

    def OnUndo(self, event):
        frame = self._get_active_frame()
        frame.OnUndo(event)

    def OnUpdateUIBackward(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIBackward'):
            frame.OnUpdateUIBackward(event)
        else:
            event.Enable(False)

    def OnUpdateUIClear(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIClear'):
            frame.OnUpdateUIClear(event)
        else:
            event.Enable(False)

    def OnUpdateUIClose(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIClose'):
            frame.OnUpdateUIClose(event)
        else:
            event.Enable(False)

    def OnUpdateUICopy(self, event):
        win = wx.Window.FindFocus()
        event.Enable(win and hasattr(win, 'CanCopy') and win.CanCopy())

    def OnUpdateUICut(self, event):
        win = wx.Window.FindFocus()
        event.Enable(win and hasattr(win, 'CanCut') and win.CanCut())

    def OnUpdateUIForward(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIForward'):
            frame.OnUpdateUIForward(event)
        else:
            event.Enable(False)

    def OnUpdateUIOpen(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIOpen'):
            frame.OnUpdateUIOpen(event)
        else:
            event.Enable(False)

    def OnUpdateUIPaste(self, event):
        win = wx.Window.FindFocus()
        event.Enable(win and hasattr(win, 'CanPaste') and win.CanPaste())

    def OnUpdateUIPlotRange(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIPlotRange'):
            frame.OnUpdateUIPlotRange(event)
        else:
            event.Check(False)
            event.Enable(False)

    def OnUpdateUIRedo(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIRedo'):
            frame.OnUpdateUIRedo(event)
        else:
            event.Enable(False)

    def OnUpdateUISave(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISave'):
            frame.OnUpdateUISave(event)
        else:
            event.Enable(False)

    def OnUpdateUISaveAs(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISaveAs'):
            frame.OnUpdateUISaveAs(event)
        else:
            event.Enable(False)

    def OnUpdateUISetLayout(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISetLayout'):
            frame.OnUpdateUISetLayout(event)
        else:
            event.Enable(False)

    def OnUpdateUISetVLim(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISetVLim'):
            frame.OnUpdateUISetVLim(event)
        else:
            event.Enable(False)

    def OnUpdateUIUndo(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIUndo'):
            frame.OnUpdateUIUndo(event)
        else:
            event.Enable(False)

    def LegacyOnQuit(self, event):
        self.ExitMainLoop()


def get_app():
    app = wx.GetApp()
    if app is None or not isinstance(app, App):
        logger.debug("Initializing Eelbrain App")
        app = App()
    return app


def run():
    app = get_app()
    if not app.IsMainLoopRunning():
        print "Starting GUI. Quit the Python application to return to the shell..."
        app.MainLoop()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>

from collections import defaultdict
from glob import glob
import itertools
import os
import re
import shutil
import subprocess

import numpy as np

from .. import fmtxt
from .._utils import LazyProperty, ui
from .._utils.com import send_email, Notifier


def _etree_expand(node, state):
    for tk, tv in node.iteritems():
        if tk == '.':
            continue

        for k, v in state.iteritems():
            name = '{%s}' % tk
            if str(v).startswith(name):
                tv[k] = {'.': v.replace(name, '')}
        if len(tv) > 1:
            _etree_expand(tv, state)


def _etree_node_repr(node, name, indent=0):
    head = ' ' * indent
    out = [(name, head + node['.'])]
    for k, v in node.iteritems():
        if k == '.':
            continue

        out.extend(_etree_node_repr(v, k, indent=indent + 3))
    return out


class LayeredDict(dict):
    """Dictionary which can store and restore states
    """
    def __init__(self, *args, **kwargs):
        self._states = []

    def __repr__(self):
        rep = ("<LayeredDict with %i stored states:\n"
               "%r>" % (len(self._states), dict.__repr__(self)))
        return rep

    def _get_state(self, index):
        "Retrieve a state (raise an appropriate error if not present)."
        if index >= len(self._states) or index < -len(self._states):
            err = ("Dict has only %i stored states, requested "
                   "%i" % (len(self._states), index))
            raise IndexError(err)
        return self._states[index]

    def get_stored(self, key, level, *args):
        """Retrieve a field value from any level

        Parameters
        ----------
        key : str
            the field name (dictionary key).
        level : int
            The level from which to retrieve the value. -1 = the current level.
        """
        state = self._get_state(level)
        return state.get(key, *args)

    def restore_state(self, index=-1, discard_tip=True):
        """Restore a previously stored state

        Parameters
        ----------
        index : int
            Index of the state which to restore (specified as index into a
            list of stored states, i.e., negative values access recently
            stored states).
        discard_tip : bool
            Discard the relevant state after restoring it. All states stored
            later are discarded either way.

        See Also
        --------
        .get_stored(): Retrieve a stored value without losing stored states
        """
        state = self._get_state(index)
        if discard_tip:
            del self._states[index:]
        elif index != -1:
            del self._states[index + 1:]

        self.clear()
        self.update(state)

    def store_state(self):
        "Store the current state"
        self._states.append(self.copy())


class _TempStateController(object):
    def __init__(self, experiment):
        self.experiment = experiment

    def __enter__(self):
        self.experiment.store_state()

    def __exit__(self, exc_type, exc_value, traceback):
        self.experiment.restore_state()


class TreeModel(object):
    """
    A hierarchical collection of format strings and field values

    Notes
    -----
    Any subclass should make sure to call the ``.store_state()` method at the
    end of initialization.
    """
    owner = None  # email address as string (for notification)

    _fmt_pattern = re.compile('\{([\w-]+)\}')

    # a dictionary of static templates (i.e., templates that do not have any hooks)
    _templates = {}
    _defaults = {}

    exclude = {}  # field_values to exclude (e.g. subjects)

    _repr_args = ()
    _repr_kwargs = ()
    def __init__(self, **state):
        self.exclude = self.exclude.copy()

        # scaffold for state
        self._fields = LayeredDict()
        self._field_values = LayeredDict()
        self._params = LayeredDict()

        # scaffold for hooks
        self._compound_members = {}
        self._compounds = defaultdict(list)
        self._eval_handlers = defaultdict(list)
        self._post_set_handlers = defaultdict(list)
        self._set_handlers = {}

        # construct initial state: make all defaults available, then set as
        # many values as we can
        self._defaults = dict(self._defaults)
        self._defaults.update(state)
        for k, v in self._templates.iteritems():
            if v is None:  # secondary field
                pass
            elif isinstance(v, basestring):
                self._register_value(k, v)
            elif isinstance(v, tuple):
                self._register_field(k, v, v[0])
            else:
                err = ("Invalid templates field value: %r. Need None, tuple "
                       "or string" % v)
                raise TypeError(err)

        if self.owner:
            task = self.__class__.__name__
            self.notification = Notifier(self.owner, task, self.show_state)

    def __repr__(self):
        args = [repr(self._fields[arg]) for arg in self._repr_args]
        kwargs = [(arg, repr(self._fields[arg])) for arg in self._repr_kwargs]

        no_initial_state = len(self._fields._states) == 0
        for k in sorted(self._fields):
            if k in self._repr_args or k in self._repr_kwargs:
                continue
            elif '{' in self._fields[k]:
                continue

            v = self._fields[k]
            if no_initial_state or v != self._fields.get_stored(k, level=0):
                kwargs.append((k, repr(v)))

        args.extend('='.join(pair) for pair in kwargs)
        args = ', '.join(args)
        return "%s(%s)" % (self.__class__.__name__, args)

    def _bind_eval(self, key, handler):
        self._eval_handlers[key].append(handler)

    def _bind_post_set(self, key, handler):
        handlers = self._post_set_handlers[key]
        if handler not in handlers:
            handlers.append(handler)

    def _bind_set(self, key, handler):
        if key in self._set_handlers:
            raise KeyError("set-handler for %r already set" % key)
        self._set_handlers[key] = handler

    def _find_missing_fields(self, add=False):
        """Find all field names occurring in field values but not as fields

        Parameters
        ----------
        add : bool
            Add non-existent field names with default value ''.
        """
        fields = set()
        for k in self._fields.keys():
            temp = self._fields[k]
            for field in self._fmt_pattern.findall(temp):
                if field not in self._fields:
                    fields.add(field)

        if add:
            for field in fields:
                self._register_value(field, '')

        return fields

    def _register_compound(self, key, elements):
        """Register a field that is composed out of other fields

        The compound always reflects ``' '.join(elements)`` including only
        elements that are not empty.

        Parameters
        ----------
        key : str
            The name of the compound field.
        elements : tuple of str
            The field names of the elements.
        """
        self._compound_members[key] = elements
        for e in elements:
            self._compounds[e].append(key)
            self._bind_post_set(e, self._update_compounds)
        self._update_compound(key)

    def _register_field(self, key, values=None, default=None, set_handler=None,
                        eval_handler=None, post_set_handler=None):
        """Register an iterable field

        Parameters
        ----------
        key : str
            Name of the field.
        values : None | sequence of str
            Possible values for this field, if known.
        default : None | str
            Set the default value (if None, the first element in values).
        set_handler : None | callable
            Function to call instead of updating the state value,
        eval_handler : None | callable
            Function to use for evaluating a value before setting.
        post_set_handler : None | callable
            Function to call after the value is changed.
        """
        if key in self._fields:
            raise KeyError("Field already exists: %r" % key)

        if set_handler is not None:
            self._bind_set(key, set_handler)
        if eval_handler is not None:
            self._bind_eval(key, eval_handler)
        if post_set_handler is not None:
            self._bind_post_set(key, post_set_handler)

        default = self._defaults.get(key, default)

        if values is not None:
            if default is None:
                if len(values):
                    default = values[0]
                else:
                    msg = "Values for %r empty, can`t set default" % key
                    raise RuntimeError(msg)
            elif default not in values:
                err = ("Default %r for %r not in values "
                       "%s" % (default, key, str(values)))
                raise ValueError(err)

            self._field_values[key] = tuple(values)

        if default is None:
            self._fields[key] = default
        else:
            kwargs = {key: default, 'add': True}
            self.set(**kwargs)

    def _register_value(self, key, default, set_handler=None,
                        eval_handler=None, post_set_handler=None):
        """Register a value with handlers

        Parameters
        ----------
        key : str
            Name of the field.
        default : None | str
            Set the default value.
        set_handler : None | callable
            Function to call instead of updating the state value,
        eval_handler : None | callable
            Function to use for evaluating a value before setting.
        post_set_handler : None | callable
            Function to call after the value is changed.
        """
        if key in self._fields:
            raise KeyError("Field already exists: %r" % key)

        if set_handler is not None:
            self._bind_set(key, set_handler)
        if eval_handler is not None:
            self._bind_eval(key, eval_handler)
        if post_set_handler is not None:
            self._bind_post_set(key, post_set_handler)

        default = self._defaults.get(key, default)
        kwargs = {key: default, 'add':True}
        self.set(**kwargs)

    def expand_template(self, temp, values=()):
        """
        Expand a template until all its subtemplates are neither in
        field names or in ``values``

        Parameters
        ----------
        values : container (implements __contains__)
            values which should not be expanded.
        """
        temp = self._fields.get(temp, temp)

        while True:
            stop = True
            for name in self._fmt_pattern.findall(temp):
                if (name in values) or (self._field_values.get(name, False)):
                    pass
                else:
                    temp = temp.replace('{%s}' % name, self._fields[name])
                    stop = False

            if stop:
                break

        return temp

    def find_keys(self, temp):
        """
        Find all terminal keys that are relevant for a template.

        Returns
        -------
        keys : set
            All terminal keys that are relevant for foormatting temp.
        """
        keys = set()
        temp = self._fields.get(temp, temp)

        for key in self._fmt_pattern.findall(temp):
            value = self._fields[key]
            if self._fmt_pattern.findall(value):
                keys = keys.union(self.find_keys(value))
            else:
                keys.add(key)

        return keys

    def format(self, string, vmatch=True, **kwargs):
        """Format a string (i.e., replace any '{xxx}' fields with their values)

        Parameters
        ----------
        string : str
            Template string.
        vmatch : bool
            For fields with known names, only allow existing field names.
        others :
            State parameters.

        Returns
        -------
        formatted_string : str
            The template temp formatted with current state values.
        """
        self.set(match=vmatch, **kwargs)

        while True:
            variables = self._fmt_pattern.findall(string)
            if variables:
                string = string.format(**self._fields)
            else:
                break

        return string

    def get(self, temp, **state):
        path = self.format('{%s}' % temp, **state)
        return path

    def get_field_values(self, field, exclude=True):
        """Find values for a field taking into account exclusion

        Parameters
        ----------
        field : str
            Field for which to find values.
        exclude : bool | list of values
            Exclude values. If True, exclude values based on ``self.exclude``.
        """
        values = self._field_values[field]
        if exclude is True:
            exclude = self.exclude.get(field, None)
        elif isinstance(exclude, basestring):
            exclude = (exclude,)

        if exclude:
            values = [v for v in values if not v in exclude]
        else:
            values = list(values)

        return values

    def iter(self, fields, exclude=True, values={}, mail=False, prog=False,
             **constants):
        """
        Cycle the experiment's state through all values on the given fields

        Parameters
        ----------
        fields : list | str
            Field(s) over which should be iterated.
        exclude : bool | dict  {str: bool, str: str, str: iterator over str}
            Exclude values from iteration. Boolean specifies whether to apply
            standard exclusion (``self.exclude``). A ``dict`` can be used to
            customize the exclusion per field with one of {field: bool,
            field: value, field: (sequence of values, )}. If only some fields
            are specified in a dict, True is assumed for absent fields.
        values : dict  {str: iterator over str}
            Fields with custom values to iterate over (instead of the
            corresponding field values) with {name: (sequence of values)}
            entries.
        group : None | str
            If iterating over subjects, use this group ('all' or a name defined
            in experiment.groups).
        prog : bool | str
            Show a progress dialog; str for dialog title.
        mail : bool | str
            Send an email when iteration is finished. Can be True or an email
            address. If True, the notification is sent to :attr:`.owner`.
        others :
            Fields with constant values throughout the iteration.
        """
        if mail is True:
            mail = self.owner

        # set constants
        self.set(**constants)
        self.store_state()

        if isinstance(fields, basestring):
            fields = [fields]
            yield_str = True
        else:
            yield_str = False
        fields = list(set(fields).difference(constants).union(values))

        # gather possible values to iterate over
        field_values = {}
        for field in fields:
            if field in values:
                field_values[field] = values[field]
            else:
                if isinstance(exclude, bool):
                    exclude_ = exclude
                elif field in exclude:
                    exclude_ = exclude[field]
                else:
                    exclude_ = True
                field_values[field] = self.get_field_values(field, exclude_)

        # pick out the fields to iterate, but drop excluded cases:
        v_lists = []
        for field in fields:
            v_lists.append(field_values[field])

        if len(v_lists):
            if prog:
                i_max = np.prod(map(len, v_lists))
                if not isinstance(prog, str):
                    prog = "MNE Experiment Iterator"
                progm = ui.progress_monitor(i_max, prog, "")
                prog = True

            for v_list in itertools.product(*v_lists):
                if prog:
                    progm.message(' | '.join(map(str, v_list)))
                self.restore_state(discard_tip=False)
                values = dict(zip(fields, v_list))
                self.set(**values)

                if yield_str:
                    yield v_list[0]
                else:
                    yield v_list

                if prog:
                    progm.advance()
        else:
            yield ()

        self.restore_state()
        if mail:
            send_email(mail, "Eelbrain Task Done", "I did as you desired, "
                       "my master.")

    def iter_temp(self, temp, exclude=True, values={}, mail=False, prog=False,
                  **constants):
        """
        Iterate through all paths conforming to a template given in ``temp``.

        Parameters
        ----------
        temp : str
            Name of a template in the MneExperiment.templates dictionary, or
            a path template with variables indicated as in ``'{var_name}'``
        """
        # if the name is an existing template, retrieve it
        keep = constants.keys() + values.keys()
        temp = self.expand_template(temp, values=keep)

        # find variables for iteration
        variables = set(self._fmt_pattern.findall(temp))

        for _ in self.iter(variables, exclude=exclude, values=values,
                           mail=mail, prog=prog, **constants):
            path = temp.format(**self._fields)
            yield path

    def restore_state(self, index=-1, discard_tip=True):
        """Restore a previously stored state

        Parameters
        ----------
        index : int
            Index of the state which to restore (specified as index into a
            list of stored states, i.e., negative values access recently
            stored states).
        discard_tip : bool
            Discard the relevant state after restoring it. All states stored
            later are discarded either way.
        """
        self._fields.restore_state(index, discard_tip)
        self._field_values.restore_state(index, discard_tip)
        self._params.restore_state(index, discard_tip)

    def set(self, match=True, allow_asterisk=False, **state):
        """Set the value of one or more fields.

        Parameters
        ----------
        match : bool
            For fields with stored values, only allow valid values.
        allow_asterisk : bool
            If a value contains '*', set the value without the normal value
            evakuation and checking mechanism.
        kwargs :
            Fields and values to set. Invalid fields raise a KeyError. Unless
            match == False, Invalid values raise a ValueError.
        """
        # fields with special set handlers
        for k in state.keys():
            if k in self._set_handlers:
                v = state.pop(k)
                self._set_handlers[k](v)

        # make sure only valid fields are set
        add = state.pop('add', False)
        if not add:
            for k in state:
                if k not in self._fields:
                    raise KeyError("No template named %r" % k)

        # eval all values
        for k in state.keys():
            eval_handlers = self._eval_handlers[k]
            if '*' in state[k] and allow_asterisk:
                pass
            elif eval_handlers:
                for handler in eval_handlers:
                    v = handler(state[k])
                    if not isinstance(v, str):
                        err = "Invalid conversion: %s=%r" % (k, v)
                        raise RuntimeError(err)
                    state[k] = v
            elif not match:
                pass
            elif k not in self._field_values:
                pass
            elif '*' in state[k]:
                pass
            elif state[k] not in self.get_field_values(k, False):
                err = ("Variable {k!r} has no value {v!r}. In order to "
                       "see valid values use e.list_values(); In order to "
                       "set a non-existent value, use e.set({k!s}={v!r}, "
                       "match=False).".format(k=k, v=state[k]))
                raise ValueError(err)

        self._fields.update(state)

        # call post_set handlers
        for k, v in state.iteritems():
            if not (allow_asterisk and '*' in v):
                for handler in self._post_set_handlers[k]:
                    handler(k, v)

    def show_fields(self, str_out=False):
        """
        Generate a table for all iterable fields and ther values.

        Parameters
        ----------
        str_out : bool
            Return the table as a string (instead of printing it).
        """
        lines = []
        for key in self._field_values:
            values = list(self.get_field_values(key))
            line = '%s:' % key
            head_len = len(line) + 1
            while values:
                v = repr(values.pop(0))
                if values:
                    v += ','
                if len(v) < 80 - head_len:
                    line += ' ' + v
                else:
                    lines.append(line)
                    line = ' ' * head_len + v

                if not values:
                    lines.append(line)

        table = os.linesep.join(lines)
        if str_out:
            return table
        else:
            print table

    def show_state(self, temp=None, empty=False, hide=[]):
        """
        List all top-level fields and their values (i.e., fields whose values
        do not contain templates).

        Parameters
        ----------
        temp : None | str
            Only show variables relevant to this template.
        empty : bool
            Show empty variables (items whose value is the empty string '').
        hide : list of str
            State variables to hide.

        Returns
        -------
        state : Table
            Table of (relevant) variables and their values.
        """
        table = fmtxt.Table('lll')
        table.cells('Key', '*', 'Value')
        table.caption('*: Value is modified from initialization state.')
        table.midrule()

        if temp is None:
            keys = (k for k, v in self._fields.iteritems()
                    if (not '{' in v) and len(v) < 60)
        else:
            keys = self.find_keys(temp)

        for k in sorted(keys):
            if k in hide:
                continue

            v = self._fields[k]
            if v != self._fields.get_stored(k, level=0):
                mod = '*'
            else:
                mod = ''

            if empty or mod or v:
                table.cells(k, mod, repr(v))

        return table

    def show_tree(self, root='root'):
        """
        Print a tree of the filehierarchy implicit in the templates

        Parameters
        ----------
        root : str
            Name of the root template (e.g., 'besa-root').
        """
        tree = {'.': root}
        root_temp = '{%s}' % root
        for k, v in self._fields.iteritems():
            if str(v).startswith(root_temp):
                tree[k] = {'.': v.replace(root_temp, '')}
        _etree_expand(tree, self._fields)
        nodes = _etree_node_repr(tree, root)
        name_len = max(len(n) for n, _ in nodes)
        path_len = max(len(p) for _, p in nodes)
        pad = ' ' * (80 - name_len - path_len)
        print os.linesep.join(n.ljust(name_len) + pad + p.ljust(path_len)
                              for n, p in nodes)

    def store_state(self):
        """Store the current state

        See also
        --------
        .restore_state() : restore a previously stored state
        """
        self._fields.store_state()
        self._field_values.store_state()
        self._params.store_state()

    @LazyProperty
    def _temporary_state(self):
        return _TempStateController(self)

    def _update_compound(self, key):
        items = []
        for item_key in self._compound_members[key]:
            value = self.get(item_key)
            if value:
                items.append(value)
        self.set(add=True, **{key: ' '.join(items)})

    def _update_compounds(self, key, _):
        for compound in self._compounds[key]:
            self._update_compound(compound)


class FileTree(TreeModel):
    """
    :class:`TreeModel` subclass for representing a file system hierarchy
    """
    _repr_args = ('root',)
    def __init__(self, **state):
        TreeModel.__init__(self, **state)
        self._make_handlers = {}
        self._cache_handlers = {}
        self._register_field('root', set_handler=self.set_root)

    def _bind_cache(self, key, handler):
        """
        The cache function is called every time the file name is retrieved and
        should recreate the file if it is outdated.
        """
        if key in self._cache_handlers:
            raise RuntimeError("Cache handler for %r already defined." % key)
        elif key in self._make_handlers:
            raise RuntimeError("Make handler for %r already defined." % key)
        self._cache_handlers[key] = handler

    def _bind_make(self, key, handler):
        """
        The make funciton is called only when the file name is retrieved and
        the file does not exist.
        """
        if key in self._cache_handlers:
            raise RuntimeError("Cache handler for %r already defined." % key)
        elif key in self._make_handlers:
            raise RuntimeError("Make handler for %r already defined." % key)
        self._make_handlers[key] = handler

    def set_root(self, root):
        root = os.path.expanduser(root)
        if not os.path.exists(root):
            raise IOError("Root does not exist: %r" % root)
        self._fields['root'] = root

    def get(self, temp, fmatch=False, vmatch=True, match=True, mkdir=False,
            make=False, **kwargs):
        """
        Retrieve a formatted template

        With match=True, '*' are expanded to match a file,
        and if there is not a unique match, an error is raised. With
        mkdir=True, the directory containing the file is created if it does not
        exist.

        Parameters
        ----------
        temp : str
            Name of the requested template.
        fmatch : bool
            "File-match": If the template contains asterisk ('*'), use glob to
            fill it in. An IOError is raised if the pattern does not match
            exactly one file.
        vmatch : bool
            "Value match": Require existence of the assigned value (only
            applies for fields with stored values).
        match : bool
            Do any matching (i.e., match=False sets fmatch as well as vmatch
            to False).
        mkdir : bool
            If the directory containing the file does not exist, create it.
        make : bool
            If a requested file does not exists, make it if possible.
        kwargs :
            Set any state values.
        """
        if not match:
            fmatch = vmatch = False

        path = TreeModel.get(self, temp, vmatch=vmatch, **kwargs)
        path = os.path.expanduser(path)

        # assert the presence of the file
        if fmatch and ('*' in path):
            paths = glob(path)
            if len(paths) == 1:
                path = paths[0]
            elif len(paths) > 1:
                err = "More than one files match %r: %r" % (path, paths)
                raise IOError(err)
            else:
                raise IOError("No file found for %r" % path)

        # create the directory
        if mkdir:
            dirname = os.path.dirname(path)
            if not os.path.exists(dirname):
                os.makedirs(dirname)

        # make the file
        if make:
            if temp in self._cache_handlers:
                self._cache_handlers[temp]()
            elif not os.path.exists(path):
                if temp in self._make_handlers:
                    with self._temporary_state:
                        self._make_handlers[temp]()
                elif temp.endswith('-dir'):
                    os.makedirs(path)
                else:
                    raise RuntimeError("No make handler for %r." % temp)

        return path

    def glob(self, temp, **state):
        """Find all files matching a certain pattern

        Parameters
        ----------
        temp : str
            Name of the path template for which to find files.

        Notes
        -----
        State parameters can include an asterisk ('*') to match multiple files.
        Uses :func:`glob.glob`.
        """
        with self._temporary_state:
            pattern = self.get(temp, allow_asterisk=True, **state)
        file_paths = glob(pattern)
        return file_paths

    def show_file_status(self, temp, row, col=None, count=True, present='X',
                         absent='-', **kwargs):
        """
        Compile a table about the existence of files

        Parameters
        ----------
        temp : str
            The name of the path template for the files to examine.
        row : str
            Field over which to alternate rows.
        col : None | str
            Field over which to alternate columns.
        count : bool
            Add a column with a number for each subject.
        present : str
            String to display when a given file is present.
        absent : str
            String to display when a given file is absent.
        others :
            ``self.iter()`` kwargs.

        Examples
        --------
        >>> e.show_file_status('raw-file', 'subject', 'raw')
             Subject   Clm   Lp40   Hp.1-lp40   Hp1-lp40
        ------------------------------------------------
         0   AD001     X     X      X           -
         1   AD002     X     X      X           -
        ...
        """
        if col is None:
            ncol = 1
        else:
            col_v = self.get_field_values(col)
            ncol = len(col_v)

        # table header
        table = fmtxt.Table('r' * bool(count) + 'l' * (ncol + 1))
        if count:
            table.cell()
        table.cell(row.capitalize())
        if col is None:
            table.cell(temp.capitalize())
        else:
            for name in col_v:
                table.cell(name.capitalize())
        table.midrule()

        # body
        for i, row_v in enumerate(self.iter(row, **kwargs)):
            if count:
                table.cell(i)
            table.cell(row_v)
            exist = []
            if col is None:
                path = self.get(temp)
                exist.append(os.path.exists(path))
            else:
                for v in col_v:
                    path = self.get(temp, **{col: v})
                    exist.append(os.path.exists(path))

            for exists in exist:
                if exists:
                    table.cell(present)
                else:
                    table.cell(absent)

        return table

    def show_file_status_mult(self, files, fields, count=True, present='X',
                              absent='-', **kwargs):
        """
        Compile a table about the existence of multiple files

        Parameters
        ----------
        files : str | list of str
            The names of the path templates whose existence to list.
        fields : str | list of str
            The names of the variables for which to list files (i.e., for each
            unique combination of ``fields``, list ``files``).
        count : bool
            Add a column with a number for each subject.
        present : str
            String to display when a given file is present.
        absent : str
            String to display when a given file is absent.

        Examples
        --------
        >>> e.show_file_status_mult(['raw-file', 'trans-file', 'fwd-file'],
        ... 'subject')
             Subject   Raw-file   Trans-file   Fwd-file
        -----------------------------------------------
         0   AD001     X          X            X
         1   AD002     X          X            X
         2   AD003     X          X            X
        ...
        """
        if not isinstance(files, (list, tuple)):
            files = [files]
        if not isinstance(fields, (list, tuple)):
            fields = [fields]

        ncol = (len(fields) + len(files))
        table = fmtxt.Table('r' * bool(count) + 'l' * ncol)
        if count:
            table.cell()
        for name in fields + files:
            table.cell(name.capitalize())
        table.midrule()

        for i, _ in enumerate(self.iter(fields, **kwargs)):
            if count:
                table.cell(i)

            for field in fields:
                table.cell(self.get(field))

            for temp in files:
                path = self.get(temp)
                if os.path.exists(path):
                    table.cell(present)
                else:
                    table.cell(absent)

        return table

    def show_in_finder(self, temp, **kwargs):
        "Reveals the file corresponding to the ``temp`` template in the Finder."
        fname = self.get(temp, **kwargs)
        subprocess.call(["open", "-R", fname])

    def push(self, dst_root, names, overwrite=False, exclude=False, **kwargs):
        """Copy files to another experiment root folder.

        Before copying any files the user is asked for confirmation.

        Parameters
        ----------
        dst_root : str
            Path to the root to which the files should be copied.
        names : str | sequence of str
            Name(s) of the template(s) of the files that should be copied.
        overwrite : bool
            What to do if the target file already exists.
        others :
            Update experiment state.

        Notes
        -----
        Use ``e.show_tree()`` to find out which element(s) to copy.
        """
        if isinstance(names, basestring):
            names = [names]

        # find files
        files = []
        for name in names:
            for src in self.iter_temp(name, exclude=exclude, **kwargs):
                if '*' in src:
                    raise NotImplementedError("Can't fnmatch here yet")

                if os.path.exists(src):
                    dst = self.get(name, root=dst_root)
                    if src == dst:
                        raise ValueError("Source == destination (%r)" % src)

                    if os.path.exists(dst):
                        flag = 'o' if overwrite else 'e'
                    else:
                        flag = ' '
                else:
                    dst = None
                    flag = 'm'
                files.append((src, dst, flag))

        # prompt for confirmation
        root = self.get('root')
        n_root = len(root)
        for src, dst, flag in files:
            if src.startswith(root):
                src = src[n_root:]
            print(' '.join((flag, src[-78:])))
        print("Flags: o=overwrite, e=skip, it exists, m=skip, source is "
              "missing")
        msg = "Proceed? (confirm with 'yes'): "
        if raw_input(msg) != 'yes':
            return

        # copy the files
        for src, dst, flag in files:
            if flag in ('e', 'm'):
                continue

            dirpath = os.path.dirname(dst)
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)

            if os.path.isdir(src):
                if flag == 'o':
                    shutil.rmtree(dst)
                shutil.copytree(src, dst)
            else:
                shutil.copy(src, dst)

    def rename(self, old, new, exclude=False):
        """
        Rename files corresponding to a pattern (or template)

        Parameters
        ----------
        old : str
            Template for the files to be renamed. Can interpret '*', but will
            raise an error in cases where more than one file fit the pattern.
        new : str
            Template for the new names.

        Examples
        --------
        The following command will collect a specific file for each subject and
        place it in a common folder:

        >>> e.rename('{root}/{subject}/info.txt',
                     '/some_other_place/{subject}s_info.txt'
        """
        new = self.expand_template(new)
        files = []
        for old_name in self.iter_temp(old, exclude):
            if '*' in old_name:
                matches = glob(old_name)
                if len(matches) == 1:
                    old_name = matches[0]
                elif len(matches) > 1:
                    err = ("Several files fit the pattern %r" % old_name)
                    raise ValueError(err)

            if os.path.exists(old_name):
                new_name = self.format(new)
                files.append((old_name, new_name))

        if not files:
            print "No files found for %r" % old
            return

        old_pf = os.path.commonprefix([pair[0] for pair in files])
        new_pf = os.path.commonprefix([pair[1] for pair in files])
        n_pf_old = len(old_pf)
        n_pf_new = len(new_pf)

        table = fmtxt.Table('lll')
        table.cells('Old', '', 'New')
        table.midrule()
        table.caption("%s -> %s" % (old_pf, new_pf))
        for old, new in files:
            table.cells(old[n_pf_old:], '->', new[n_pf_new:])

        print table

        msg = "Rename %s files (confirm with 'yes')? " % len(files)
        if raw_input(msg) == 'yes':
            for old, new in files:
                dirname = os.path.dirname(new)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                os.rename(old, new)

    def rename_field(self, temp, field, old, new, exclude=False, **kwargs):
        """Change the value of one field in paths corresponding to a template

        Parameters
        ----------
        temp : str
            Template name.
        field : str
            Field to change.
        old : str
            Old value.
        new : str
            New value.
        kwargs :
            ``self.iter_temp`` arguments.
        """
        items = []  # (tag, src, dst)
        kwargs[field] = old
        dst_kwa = {field: new}
        for src in self.iter_temp(temp, exclude, ** kwargs):
            dst = self.get(temp, **dst_kwa)
            if os.path.exists(src):
                if os.path.exists(dst):
                    tag = 'o'
                else:
                    tag = ' '
            else:
                tag = 'm'
            items.append((tag, src, dst))

        src_prefix = os.path.commonprefix(tuple(item[1] for item in items))
        dst_prefix = os.path.commonprefix(tuple(item[2] for item in items))
        src_crop = len(src_prefix)
        dst_crop = len(dst_prefix)

        # print info
        if src_prefix == dst_prefix:
            lines = ['in ' + src_prefix, '']
        else:
            lines = [src_prefix, '->' + dst_prefix, '']

        for tag, src, dst in items:
            lines.append('%s %s -> %s' % (tag, src[src_crop:], dst[dst_crop:]))
        lines.append('')
        msg = 'Legend  m: source is missing;  o: will overwite a file'
        lines.append(msg)
        print os.linesep.join(lines)
        rename = tuple(item for item in items if item[0] == ' ')
        if not rename:
            return

        msg = "Rename %i files (confirm with 'yes')? " % len(rename)
        if raw_input(msg) != 'yes':
            return

        for _, src, dst in rename:
            os.rename(src, dst)
        print "Done"

    def rm(self, temp, exclude=False, values={}, v=False, **constants):
        """
        Remove all files corresponding to a template

        Asks for confirmation before deleting anything. Uses glob, so
        individual templates can be set to '*'.

        Parameters
        ----------
        temp : str
            The template.
        exclude : bool | dict
            Exclude specific field values.
        values : dict
            Provide specific values by field.
        v : bool
            Verbose mode (print all filename patterns that are searched).
        others :
            Set fields.
        """
        files = []
        for fname in self.iter_temp(temp, exclude, values=values, **constants):
            fnames = glob(fname)
            if v:
                print "%s -> %i" % (fname, len(fnames))
            if fnames:
                files.extend(fnames)
            elif os.path.exists(fname):
                files.append(fname)

        if files:
            root = self.get('root')
            print "root: %s\n" % root
            root_len = len(root)
            for name in files:
                if name.startswith(root):
                    print name[root_len:]
                else:
                    print name
            msg = "Delete %i files (confirm with 'yes')? " % len(files)
            if raw_input(msg) == 'yes':
                for path in files:
                    if os.path.isdir(path):
                        shutil.rmtree(path)
                    else:
                        os.remove(path)
        else:
            print "No files found for %r" % temp




"""
Plot multidimensional uniform time series.
"""

from __future__ import division

import logging
import math

import numpy as np
import matplotlib.pyplot as plt

from .. import _data_obj as _dta
from . import _base


class _plt_im_array(object):
    def __init__(self, ax, ndvar, overlay, dims=('time', 'sensor'),
                 extent=None, interpolation=None, vlims={}, cmaps={},
                 contours={}):
        im_kwa = _base.find_im_args(ndvar, overlay, vlims, cmaps)
        self._contours = _base.find_ct_args(ndvar, overlay, contours)
        self._meas = ndvar.info.get('meas', _base.default_meas)
        self._dims = dims

        data = self._data_from_ndvar(ndvar)
        if im_kwa is not None:
            self.im = ax.imshow(data, origin='lower', aspect='auto',
                                extent=extent, interpolation=interpolation,
                                **im_kwa)
            self._cmap = im_kwa['cmap']
        else:
            self.im = None

        # store attributes
        self.ax = ax
        self.cont = None
        self._data = data
        self._aspect = 'auto'
        self._extent = extent

        # draw flexible part
        self._draw_contours()

    def _data_from_ndvar(self, ndvar):
        data = ndvar.get_data(self._dims)
        if data.ndim > 2:
            assert data.shape[0] == 1
            data = data[0]
        return data

    def _draw_contours(self):
        if self.cont:
            for c in self.cont.collections:
                c.remove()
            self.cont = None

        if self._contours:
            levels = sorted(self._contours)
            colors = [self._contours[l] for l in levels]
            self.cont = self.ax.contour(self._data, levels=levels,
                                        colors=colors, aspect=self._aspect,
                                        origin='lower', extent=self._extent)

    def add_contour(self, meas, level, color):
        if self._meas == meas:
            self._contours[level] = color
            self._draw_contours()

    def get_kwargs(self):
        "Get the arguments required for plotting the im"
        if self.im:
            vmin, vmax = self.im.get_clim()
            args = dict(cmap=self._cmap, vmin=vmin, vmax=vmax)
        else:
            args = {}
        return args

    def set_cmap(self, cmap, meas=None):
        if (self.im is not None) and (meas is None or meas == self._meas):
            self.im.set_cmap(cmap)
            self._cmap = cmap

    def set_data(self, ndvar):
        data = self._data_from_ndvar(ndvar)
        if self.im is not None:
            self.im.set_data(data)
        self._data = data
        self._draw_contours()

    def set_vlim(self, vmax, meas, vmin):
        if self.im is None:
            return
        elif (meas is not None) and (self._meas != meas):
            return

        if vmax is None:
            _, vmax = self.im.get_clim()

        vmin, vmax = _base.fix_vlim_for_cmap(vmin, vmax, self._cmap)
        self.im.set_clim(vmin, vmax)


class _ax_im_array(object):
    def __init__(self, ax, layers, x='time', xlabel=True, ylabel=True,
                 title=None, tick_spacing=0.1, interpolation=None, vlims={},
                 cmaps={}, contours={}):
        """
        plots segment data as im

        define a colorspace by supplying one of those kwargs: ``colorspace`` OR
        ``p`` OR ``vmax``

        """
        self.ax = ax
        self.data = layers
        self.layers = []
        epoch = layers[0]

        xdim = epoch.get_dim(x)
        if epoch.ndim == 2:
            xdim_i = epoch.dimnames.index(x)
            ydim_i = {1:0, 0:1}[xdim_i]
            y = epoch.dimnames[ydim_i]
        else:
            err = ("Need 2 dimensions, got %i" % epoch.ndim)
            raise ValueError(err)

        ydim = epoch.get_dim(y)
        if not ydim.adjacent:
            ydim = _dta.Var(np.arange(len(ydim)), y)

        # plot
        overlay = False
        extent = (xdim[0], xdim[-1], ydim[0], ydim[-1])
        for l in layers:
            p = _plt_im_array(ax, l, overlay, (y, x), extent, interpolation,
                              vlims, cmaps, contours)
            self.layers.append(p)
            overlay = True

        xlabel = _base._axlabel(x, xlabel)
        if xlabel:
            ax.set_xlabel(xlabel)

        if ylabel:
            if ylabel is True:
                ylabel = ydim.name
            ax.set_ylabel(ylabel)

        # x-ticks
        tickstart = math.ceil(xdim[0] / tick_spacing) * tick_spacing
        tickstop = xdim[-1] + tick_spacing / 1e4
        if tickstop - tickstart > tick_spacing:
            ticks = np.arange(tickstart, tickstop, tick_spacing)
            if len(ticks) > 2 and ticks[0] < 0 and ticks[0] == xdim[0]:
                ticks = ticks[1:]
        else:
            ticks = np.array([xdim[0], xdim[-1]])
        ax.xaxis.set_ticks(ticks)
        ticklabels = _base._ticklabels(ticks, xdim.name)
        ax.xaxis.set_ticklabels(ticklabels)
        if xdim.name == 'time':
            ax.x_fmt = "t = %.3f s"

        # y-ticks
        if y == 'sensor':  # make sure y-ticklabels are all integers
            locs = ax.yaxis.get_ticklocs()
            if any(locs != locs.round()):
                idx = np.where(locs == locs.round())[0]
                locs = locs[idx]
                labels = map(lambda x: str(int(x)), locs)
                ax.yaxis.set_ticks(locs)
                ax.yaxis.set_ticklabels(labels)

        # title
        if title is None:
            title = _base.str2tex(epoch.name)
        ax.set_title(title)

    def add_contour(self, meas, level, color):
        for l in self.layers:
            l.add_contour(meas, level, color)

    def set_cmap(self, cmap, meas=None):
        """Change the colormap in the array plot

        Parameters
        ----------
        cmap : str | colormap
            New colormap.
        meas : None | str
            Measurement to which to apply the colormap. With None, it is
            applied to all.
        """
        for l in self.layers:
            l.set_cmap(cmap, meas)

    def set_data(self, layers):
        self.data = layers
        for l, p in zip(layers, self.layers):
            p.set_data(l)

    def set_vlim(self, vmax=None, meas=None, vmin=None):
        for l in self.layers:
            l.set_vlim(vmax, meas, vmin)


class Array(_base._EelFigure):
    def __init__(self, epochs, Xax=None, title=None, xlabel=True, ylabel=True,
                 ds=None, **layout):
        """
        Plot uts data to a rectangular grid.

        Parameters
        ----------
        epochs : NDVar
            If data has only 1 dimension, the x-axis defines epochs.
        Xax : None | categorial
            Create a separate plot for each cell in this model.
        title : None | string
            Figure title.
        xlabel, ylabel : bool | str
            I True, determine from the data.
        ds : None | Dataset
            If a Dataset is provided, ``epochs`` and ``Xax`` can be specified
            as strings.
        """
        epochs = _base.unpack_epochs_arg(epochs, 2, Xax, ds)

        nax = len(epochs)
        _base._EelFigure.__init__(self, "Array Plot", nax, layout,
                                 figtitle=title)

        self.plots = []
        vlims = _base.find_fig_vlims(epochs)
        for i, ax, layers in zip(xrange(nax), self._axes, epochs):
            _ylabel = ylabel if i == 1 else None
            _xlabel = xlabel if i == nax - 1 else None
            p = _ax_im_array(ax, layers, xlabel=_xlabel, ylabel=ylabel,
                             vlims=vlims)
            self.plots.append(p)

        self._show()

    def add_contour(self, meas, level, color='k'):
        """Add a contour line

        Parameters
        ----------
        meas : str
            The measurement for which to add a contour line.
        level : scalar
            The value at which to draw the contour.
        color : matplotlib color
            The color of the contour line.
        """
        for p in self.plots:
            p.add_contour(meas, level, color)
        self.draw()

    def set_cmap(self, cmap, meas=None):
        """Change the colormap in the array plots

        Parameters
        ----------
        cmap : str | colormap
            New colormap.
        meas : None | str
            Measurement to which to apply the colormap. With None, it is
            applied to all.
        """
        for p in self.plots:
            p.set_cmap(cmap, meas)
        self.draw()

    def set_vlim(self, vmax=None, meas=None, vmin=None):
        for p in self.plots:
            p.set_vlim(vmax, meas, vmin)
        self.draw()


def _t_axes(rect, xmin=-.5, xmax=1., vmax=1.5, vbar=1., markevery=.5, ticks=False):
    """
    creates and returns a new t-axes using rect

    vbar: extent of the vertical bar at the origin (+vbar to -vbar)

    """
    ax = plt.axes(rect, frameon=False)
    ax.set_axis_off()

    # vertical bar
    plt.plot([0, 0], [-vbar, vbar], 'k',
           marker='_', mec='k', mew=1, mfc='k')

    # horizontal bar
    xdata = np.arange(-markevery, xmax + 1e-5, markevery)
    xdata[0] = xmin
    plt.plot(xdata, xdata * 0, 'k', marker='|', mec='k', mew=1, mfc='k')
    logging.debug("xdata: %s" % str(xdata))

    # labels
    if ticks:
        ax.text(xmax, vbar * .1, r"$%s s$" % xmax, verticalalignment='bottom', horizontalalignment='center')
        ax.text(-vbar * .05, vbar, r"$%s \mu V$" % vbar, verticalalignment='center', horizontalalignment='right')
        ax.text(-vbar * .05, -vbar, r"$-%s \mu V$" % vbar, verticalalignment='center', horizontalalignment='right')
    ax.set_ylim(-vmax, vmax)
    ax.set_xlim(xmin, xmax)
    return ax



def _axgrid_sensors(sensorLocs2d, figsize=(8, 8),
                    spacing=.2, frame=.01, figvstretch=1,
                    header=0, footer=0,  # in inches
                    axes_legend_loc='lower left', **axkwargs):
    """
    creates topographocally distributed t-axes

     returns
        - list of t-axes
        - axes-legend (indicating t and V) (or None if axes_legend_loc == False)
    """
    # determine figure size and create figure
    sensorLocs2d[:, 1] *= figvstretch
    x, y = sensorLocs2d.max(axis=0) - sensorLocs2d.min(axis=0)
    ratio = (x + spacing) / (y + spacing)
    x_size, y_size = figsize
    if x_size == -1:
        x_size = y_size * ratio
    elif y_size == -1:
        y_size = x_size / ratio
    # take into account footer & header
    y_size += footer + header
    relative_footer = footer / float(y_size)
    relative_header = header / float(y_size)
    logging.debug(" _axgrid_sensors determined figsize %s x %s" % (x_size, y_size))
    fig = plt.figure(figsize=(x_size, y_size))
    # determine axes locations
    locs = sensorLocs2d
    # normalize range to 0--1
    locs -= locs.min(0)
    locs /= locs.max(0)
    # locs--> lower left points of axes
    locs[:, 0] *= (1 - spacing - 2 * frame)
    locs[:, 0] += frame
    locs[:, 1] *= (1 - spacing - 2 * frame - relative_header - relative_footer)
    locs[:, 1] += frame + relative_footer
    # print locs
    # add axes
    axes = [ _t_axes([x, y, spacing, spacing], **axkwargs) for x, y in locs ]
    if axes_legend_loc:
        x, y = _base._loc(axes_legend_loc, size=(1.1 * spacing, spacing), frame=frame)
        axes_legend = _t_axes([x, y, spacing, spacing], ticks=True, **axkwargs)
    else:
        axes_legend = None
    return axes, axes_legend


class _plt_utsnd:

    def __init__(self, ax, epoch, sensors=None, *args, **kwargs):
        """
        uts plot for a single epoch

        Parameters
        ----------
        ax : matplotlib axes
            Target axes.
        epoch : NDVar (sensor by time)
            Epoch to plot.
        sensors : None | True | numpy index
            The sensors to plot (None or True -> all sensors).
        others :
            Matplotlib plot() arguments.
        """
        if sensors is not None and sensors is not True:
            epoch = epoch.sub(sensor=sensors)

        Y = epoch.get_data(('time', 'sensor'))
        x = epoch.time.x
        kwargs['label'] = epoch.name
        handles = ax.plot(x, Y, *args, **kwargs)

        for y, kwa in _base.find_uts_hlines(epoch):
            ax.axhline(y, **kwa)

        self.x = x
        self.lines = handles

    def remove(self):
        while self.lines:
            self.lines.pop().remove()


class _plt_extrema:

    def __init__(self, ax, epoch, **plot_kwargs):
        data = epoch.get_data(('time', 'sensor'))
        Ymin = data.min(1)
        Ymax = data.max(1)
        x = epoch.time.x

        handle = ax.fill_between(x, Ymin, Ymax, **plot_kwargs)
        ax.set_xlim(x[0], x[-1])

        self.x = x
        self.fill = handle

    def remove(self):
        self.fill.remove()


class _ax_butterfly(object):

    def __init__(self, ax, layers, sensors=None, extrema=False, title='{name}',
                 xlabel=True, ylabel=True, color=None, vlims={}):
        """
        Parameters
        ----------
        vmin, vmax: None | scalar
            Y axis limits.
        """
        self.ax = ax
        self.data = layers
        self.layers = []
        self._xvalues = []  # values on the x axis
        self._meas = None

        vmin, vmax = _base.find_uts_ax_vlim(layers, vlims)

        name = ''
        overlay = False
        for l in layers:
            uts_args = _base.find_uts_args(l, overlay, color)
            if uts_args is None:
                continue
            overlay = True

            # plot
            if extrema:
                h = _plt_extrema(ax, l, **uts_args)
            else:
                h = _plt_utsnd(ax, l, sensors, **uts_args)

            self.layers.append(h)
            if not name:
                name = getattr(l, 'name', '')

            self._xvalues = np.union1d(self._xvalues, l.time.x)

        # axes decoration
        ax.set_xlim(min(l.x[0] for l in self.layers),
                    max(l.x[-1] for l in self.layers))

        ticks = ax.xaxis.get_ticklocs()
        ticklabels = _base._ticklabels(ticks, 'time')
        ax.xaxis.set_ticklabels(ticklabels)

        xlabel = _base._axlabel('time', xlabel)
        if xlabel:
            ax.set_xlabel(xlabel)

        l = layers[0]
        if ylabel is True:
            ylabel = l.info.get('unit', None)
        if ylabel:
            ax.set_ylabel(ylabel)

    #    ax.yaxis.set_offset_position('right')
        ax.yaxis.offsetText.set_va('top')

        ax.x_fmt = "t = %.3f s"
        if isinstance(title, str):
            ax.set_title(title.format(name=name))

        self.set_vlim(vmax, vmin)

    def set_vlim(self, vmax=None, vmin=None):
        if vmin is None and vmax is not None:
            vmin = -vmax
        self.ax.set_ylim(vmin, vmax)
        vmin, vmax = self.ax.get_ylim()
        self.vmin = vmin
        self.vmax = vmax


class Butterfly(_base._EelFigure):
    "Plot data in a butterfly plot."
    def __init__(self, epochs, Xax=None, sensors=None, title=None,
                 axtitle='{name}', xlabel=True, ylabel=True, color=None,
                 ds=None, **layout):
        """
        Parameters
        ----------
        epochs : (list of) NDVar
            Data to plot.
        Xax : None | categorial
            Create a separate plot for each cell in this model.
        sensors: None or list of sensor IDs
            sensors to plot (``None`` = all)
        title : None | string
            Figure title.
        color : matplotlib color
            default (``None``): use segment color if available, otherwise
            black; ``True``: alternate colors (mpl default)
        """
        epochs = _base.unpack_epochs_arg(epochs, 2, Xax, ds)
        super(Butterfly, self).__init__('Butterfly Plot', len(epochs), layout,
                                        figtitle=title)

        self.plots = []
        vlims = _base.find_fig_vlims(epochs, True)
        for ax, layers in zip(self._axes, epochs):
            h = _ax_butterfly(ax, layers, sensors=sensors, vlims=vlims,
                              title=axtitle, xlabel=xlabel, ylabel=ylabel,
                              color=color)
            self.plots.append(h)
            xlabel = None

        self._show()

    def set_vlim(self, vmax=None, vmin=None):
        for p in self.plots:
            p.set_vlim(vmax, vmin)
        self.draw()


class _ax_bfly_epoch:
    def __init__(self, ax, epoch, xlabel=True, ylabel=True, plot_range=True,
                 traces=None, color=None, lw=0.2, mark=None, mcolor='r',
                 mlw=0.8, antialiased=True, state=True, vlims={}):
        """Specific plot for showing a single sensor by time epoch

        Parameters
        ----------
        ...
        epoch : NDVar
            Sensor by time epoch.
        ...
        lw : scalar
            Sensor trace plot Line width (default 0.5).
        mlw : scalar
            Marked sensor plot line width (default 1).
        """
        if color is None:
            if traces is None:
                color = '0.5'
            else:
                color = 'k'

        self.ax = ax
        self._traces = None
        self._marked_traces = None
        self._range = None
        self._state_h = []

        self._trace_kwargs = dict(color=color, lw=lw, antialiased=antialiased)
        self._range_kwargs = dict(color=color, antialiased=antialiased)
        self._mark_kwargs = dict(color=mcolor, lw=mlw, antialiased=antialiased,
                                 sensors=mark)
        self._do_plot_range = plot_range
        if traces is True:
            self._do_plot_traces = True
        elif traces is None or traces is False:
            self._do_plot_traces = False
        else:
            self._do_plot_traces = True
            self._trace_kwargs['sensors'] = traces

        self._tmin = epoch.time[0]
        self._tmax = epoch.time[-1]
        self._ylim = _base.find_uts_ax_vlim([epoch], vlims)

        self.ax.x_fmt = "t = %.3f s"

        # ax decoration
        xlabel = _base._axlabel('time', xlabel)
        if xlabel:
            ax.set_xlabel(xlabel)

        if ylabel is True:
            ylabel = epoch.info.get('unit', None)
        if ylabel:
            self.ax.set_ylabel(ylabel)
            self.ax.yaxis.offsetText.set_va('top')

        # create initial plots
        self.set_data(epoch)
        self.set_state(state)

    def _plot_range(self):
        "plot the range between sensors"
        self._rm_range()
        if not self._do_plot_range:
            return

        self._range = _plt_extrema(self.ax, self.epoch, **self._range_kwargs)

    def _plot_traces(self):
        "Plot traces for individual sensors"
        self._rm_traces()
        if self._do_plot_traces:
            h = _plt_utsnd(self.ax, self.epoch, **self._trace_kwargs)
            self._traces = h
        if self._mark_kwargs['sensors']:
            h = _plt_utsnd(self.ax, self.epoch, **self._mark_kwargs)
            self._marked_traces = h

    def _rm_range(self):
        "Remove the range from the plot"
        if self._range:
            self._range.remove()
            self._range = None

    def _rm_traces(self):
        "Remove the traces from the plot"
        if self._traces:
            self._traces.remove()
            self._traces = None
        if self._marked_traces:
            self._marked_traces.remove()
            self._marked_traces = None

    def set_data(self, epoch):
        self.epoch = epoch
        self._plot_range()
        self._plot_traces()
        self.set_ax_lim()

    def set_ax_lim(self):
        self.ax.set_xlim(self._tmin, self._tmax)
        ylim = self._ylim
        if ylim:
            if np.isscalar(ylim):
                self.ax.set_ylim(-ylim, ylim)
            else:
                y_min, y_max = ylim
                self.ax.set_ylim(y_min, y_max)

    def set_state(self, state):
        "Set the state (True=accept / False=reject)"
        if state:
            while self._state_h:
                h = self._state_h.pop()
                h.remove()
        else:
            if not self._state_h:
                h1 = self.ax.plot([0, 1], [0, 1], color='r', linewidth=1,
                                  transform=self.ax.transAxes)
                h2 = self.ax.plot([0, 1], [1, 0], color='r', linewidth=1,
                                  transform=self.ax.transAxes)
                self._state_h.extend(h1 + h2)

    def set_ylim(self, ylim):
        self._ylim = ylim
        self.set_ax_lim()
